''' Represents a sub or a master calibration image; handles all 
    FITs reading/writing plus metadata from FITs header/filename parsing
'''

import os
import time
import glob
import numpy as np
from datetime import datetime
from astropy.io import fits
from loguru import logger

from jocular import __version__
from jocular.exposurechooser import str_to_exp

''' map from FITs names (converted to lower case) to Image attributes; 
    (1) can have multiple names mapping to same attribute
    (2) keys must be lower case for lookup, even though UC in FITs
    (3) values must have case as used in Image props and rest of program
    (4) any that are not same in UC must be specified in imagepropmap
'''
fitspropmap = {
        'exptime': 'exposure', 'expo': 'exposure', 'exp': 'exposure',
        'filt': 'filter',
        'roi_x': 'ROI_x', 'roi_y': 'ROI_y', 'roi_w': 'ROI_w', 'roi_h': 'ROI_h',
        'calibrat': 'calibration_method',
        'subtype': 'sub_type', 'imagetyp': 'sub_type',
        'temperat': 'temperature', 'temp': 'temperature', 'ccd-temp': 'temperature',
        'stackcnt': 'nsubs',
        'xpixsz': 'pixel_width', 'ypixsz': 'pixel_height',
        'xbinning': 'binning',
        'ybinning': 'binning',
        'instrume': 'camera'
        }

''' used to map from image prop names to fits name if they are difference; 
    can't use inverse of fitspropmap as it is not unique
'''
imagepropsmap = {
    'calibration_method': 'calibrat',
    'pixel_width': 'xpixsz',
    'pixel_height': 'ypixsz',
    'temperature': 'temperat'
}

fitscomments = {
    'calibration_method': 'calibration method applied to master flats'
}

''' properties we must have (even if None) in each Image instance
    these are overridden by name and FITs properties
'''
default_props = {
    'exposure': None, 
    'sub_type': 'light', 
    'filter': 'L', 
    'temperature': None,
    'gain': None,
    'offset': None,
    'camera': None,
    'binning': 1,
    'ROI_x': None,
    'ROI_y': None,
    'ROI_w': None,
    'ROI_h': None,
    'pixel_width': None,
    'pixel_height': None,
    'calibration_method': 'None',
    'nsubs': None}

# map from possible found filter names to the Jocular scheme
filter_map = {'r': 'R', 'g': 'G', 'b': 'B', 'red': 'R', 'green': 'G', 
    'blue': 'B', 'dark': 'dark',  
    'ha': 'H', 'halpha': 'H', 'h': 'H',
    's': 'S', 'sii': 'S', 's': 'S',
    'oiii': 'O', 'o': 'O',
    'spect': 'spec', 'l': 'L', 'lum': 'L', 'no': 'L', 'none': 'L'}


# this should be moved to a fits module
def update_wcs(im=None, path=None, w=None, h=None, fov_w=None, fov_h=None, ra0=None, dec0=None, north=None):
    ''' add or update FITs at path with WCS keys
        see 
            [1] https://fits.gsfc.nasa.gov/standard40/fits_standard40aa-le.pdf
            [2] https://learn.astropy.org/tutorials/celestial_coords1.html
            [3] https://docs.astropy.org/en/stable/api/astropy.wcs.Wcsprm.html#astropy.wcs.Wcsprm
    '''
    from astropy.wcs import WCS

    # provide rotation matrix
    theta = np.radians(north)
    costheta, sintheta = np.cos(theta), np.sin(theta)

    # NB axis 1 is RA
    wcs_dict = {
        'CTYPE1': 'RA---TAN', # RA to gnomonic
        'CUNIT1': 'deg',      # RA in degrees
        'CDELT1': -fov_w / w, # degrees increment per pixel (neg because RA decreases to R)
        'CRPIX1': w // 2,     # central pixel in x; might need to alter this by 1 
        'CRVAL1': ra0,        # RA value at central pixel
        'NAXIS1': w,          # size of axis 1
        'CTYPE2': 'DEC--TAN', 
        'CUNIT2': 'deg', 
        'CDELT2': fov_h / h, 
        'CRPIX2': h // 2,     # might need to alter this by 1?
        'CRVAL2': dec0,
        'PC1_1':  costheta,   # rotation matrix
        'PC1_2':  -sintheta,  #   cos -sin  ===   PC1_1 PC1_2
        'PC2_1':  sintheta,   #   sin  cos  ===   PC2_1 PC2_2
        'PC2_2':  costheta,   # 
        'NAXIS2': h
    }

    wcs_hdr = WCS(wcs_dict)

    # # to do: update FITs header
    # # to start with, generate a testfits with the image and WCS info
    # # to plot and overlay axes
    from astropy.io import fits
    from datetime import datetime

    hdu = fits.PrimaryHDU(im, header=wcs_hdr.to_header())
    hdr = hdu.header
    hdr['CREATOR'] = ('Jocular')
    date_obs = datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
    hdr['DATE-OBS'] = date_obs
    hdu.writeto('platesolved_wcs_' + date_obs + '.fits')


def convert_fits_props(hdr):
    ''' form dict from Image prop to value but don't include unknown FITs props
    '''
    props = {}
    for k, v in hdr.items():
        ktran = fitspropmap.get(k.lower(), k.lower())
        if ktran in default_props:
            props[ktran] = v
    return props


def is_fit(f):
    return f.lower().endswith('.fit') or f.lower().endswith('.fits')



class ImageNotReadyException(Exception):
    pass


def fits_in_dir(path):
    # get all fits in path, returning sorted by modification time
    fits = [f for f in glob.glob(os.path.join(path, '*')) if is_fit(f)]
    fits = {f: os.path.getmtime(f) for f in fits}
    return sorted(fits, key=fits.get)


def save_image(data=None, path=None, capture_props=None):
    ''' saves sub or master to path
        capture_props is a dict to save e.g. gain, offset, expo
    '''

    dupes = [
        ('SUBTYPE', 'IMAGETYP'),
        ('EXPOSURE', 'EXPTIME'), 
        ('CAMERA', 'INSTRUME'), 
        ('NSUBS', 'STACKCNT'),
        ('BINNING', 'XBINNING'),
        ('BINNING', 'YBINNING'),
        ('TEMPERAT', 'CCD-TEMP')
    ]

    exposure = capture_props.get('exposure', None)
    sub_type = capture_props.get('sub_type', 'L')
    filt = capture_props.get('filter', 'L')
    temperature = capture_props.get('temperature', None)
    nsubs = capture_props.get('nsubs', None)

    try:
        hdu = fits.PrimaryHDU(data)
        hdr = hdu.header
        hdr['CREATOR'] = ('Jocular', __version__)
        hdr['DATE-OBS'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]
        if exposure is not None:
            hdr['EXPOSURE'] = (exposure, 'seconds')
        if sub_type in {'master dark', 'master bias'}:
            hdr['FILTER'] = 'dark'
        else:
            hdr['FILTER'] = filt
        hdr['SUBTYPE'] = sub_type
        if temperature is not None:
            hdr['TEMPERAT'] = temperature
        if nsubs is not None:
            hdr['NSUBS'] = nsubs 

        # add any other keys that are not in the header
        # ensuring that any conversions are done
        for k, v in capture_props.items():
            newk = imagepropsmap.get(k.lower(), k).upper()
            if v is not None and newk not in hdr:
                hdr[newk] = (v, fitscomments.get(k, ''))

        # duplicate values for alternative keys
        for key, altkey in dupes:
            if key in hdr:
                hdr[altkey] = hdr[key]

        hdu.writeto(path, overwrite=True)
        logger.info(f'saved {path}')
        for k, v in hdr.items():
            logger.trace(f'{k:9s} = {v}')

    except Exception as e:
        logger.warning(f'unable to save to {path} ({e})')

# def update_fits_header(path, exposure=None, sub_type=None, temperature=None):
#     ''' modify headers; not used currently but will be used when advanced
#         save is re-implemented in ObjectIO
#     '''
#     try:
#         with fits.open(path, mode='update') as hdu:
#             hdr = hdu[0].header
#             if exposure is not None:
#                 hdr['EXPOSURE'] = (exposure, 'seconds')
#             if sub_type is not None:
#                 hdr['SUBTYPE'] = sub_type
#             if temperature is not None:
#                 hdr['TEMPERAT'] = temperature
#             hdu.close()
#     except Exception as e:
#         logger.exception('Unable to read fits header for {:} ({:})'.format(path, e))
#         return


class Image:
    ''' Represents sub or master
    '''

    def __init__(self, path=None, verbose=False, check_image_data=False):
        ''' Generate Image instance from path. Throw ImageNotReadyException 
            in cases where re-reading on next event cycle might be successful.
        '''

        # while testing
        verbose = False

        if path is None:
            return

        if not is_fit(path):
            raise Exception(f'not a fits file {path}')

        try:
            # added in v0.5.6 ignore missing end
            with fits.open(path, ignore_missing_end=True) as hdu1:
                hdu1.verify('silentfix')
                hdr = hdu1[0].header
                # we check that the image is readable if necessary
                if check_image_data:
                    bp = hdr['BITPIX']
                    if bp < 0:
                        self.image = hdu1[0].data
                    else:
                        self.image = hdu1[0].data / 2 ** bp

                    self.minval = np.min(self.image) * 100
                    self.maxval = np.percentile(self.image, 99.99) * 100
                    self.meanval = np.mean(self.image) * 100
                    self.overexp = np.mean(self.image > .99) * 100
                else:
                    self.image = None
                    self.minval, self.maxval, self.meanval, self.overexp = .0, .0, .0, .0

        except Exception as e:
            self.image = None
            raise ImageNotReadyException(f'Cannot read fits header for {path} ({e})')

        if verbose:
            logger.trace('From FITS header')
            for k, v in hdr.items():
                logger.trace(f'{k:9s} = {v}')

        # check for a well-formed image (new in v0.5)
        if self.image is not None:
            shape = self.image.shape
            if len(shape) == 3:
                if shape[0] != 3 and shape[2] != 3:
                    raise Exception(f'3D fits without 3 planes {path}')
            elif len(shape) != 2:
                raise Exception(f'can only handle 2D or 3D fits {path}')

        try:
            self.path = path
            self.create_time = datetime.fromtimestamp(os.path.getmtime(path))
            self.arrival_time = datetime.now()
            self.age = (self.arrival_time - self.create_time).days
            self.created_by_jocular = 'CREATOR' in hdr and hdr['CREATOR'].startswith('Jocular')
            self.shape = hdr['NAXIS1'], hdr['NAXIS2']
            self.shape_str = f'{self.shape[0]}x{self.shape[1]}'
            self.fullname = os.path.basename(path)
            self.name = os.path.splitext(os.path.basename(path).lower())[0]
            self.pp_create_time = self.create_time.strftime('%d %b %y %H:%M:%S')
        except Exception as e:
            logger.exception(e)

        # extract any relevant information from FITs header
        fits_props = convert_fits_props(hdr)

        try:
            # is it a master?
            self.is_master = self.name.startswith('master') or \
                fits_props.get('sub_type', '').lower().startswith('master') # or \
                # fits_props.get('nsubs', 0) > 1

        except Exception as e:
            logger.exception(e)

        # get properties from name
        if self.is_master:
            name_props = self.parse_master(self.name[6:])
        else:
            name_props = self.parse_sub(self.name)

        # form properties by overriding
        props = {**default_props, **name_props, **fits_props}

        # rationalise properties
        try:
            if props['exposure'] is not None:
                try:
                    props['exposure'] = str_to_exp(props['exposure'])
                except:
                    props['exposure'] = None

            st = props['sub_type'].lower()
            if st.startswith('master'):  # change 'master ' to 'master'
                st = st[6:].strip()
            if st[:4] in {'bias', 'dark', 'flat'}:
                props['sub_type'] = st[:4]
            else:
                props['sub_type'] = 'light'

            # filter might be R, r, red, filter: red, red filter, ...
            if type(props['filter']) == str:
                v = props['filter'].lower()
                if v.endswith('filter'):
                    v = v[:-6].strip()
                if v.startswith('filter'):
                    v = v[6:].strip()
                props['filter'] = filter_map.get(v, v)   # untested change in v0.5
            else:
                props['filter'] = 'L'

            # temperature format might be a number, or followed by C
            if props['temperature'] is not None:
                v = str(props['temperature']).lower()
                if v.endswith('c'):
                    v = v[:-1].strip()
                try:
                    props['temperature'] = float(v)
                except:
                    props['temperature'] = None

            # number of subs
            if props['nsubs'] is not None:
                try:
                    props['nsubs'] = int(props['nsubs'])
                except:
                    props['nsubs'] = None

        except Exception as e:
            logger.exception(e)

        # assign properties to Image instance
        for p in default_props:
            setattr(self, p, props[p])

        self.status = 'select'
        self.aligned = 'no'
        self.arrival_time = int(time.time())
        self.keyframe = False
        self.calibrations = {'dark': False, 'flat': False, 'bias': False}

        self.describe(
            fits_props=fits_props, 
            name_props=name_props, 
            verbose=verbose)


    def describe(self, fits_props=None, name_props=None, verbose=False):
        created_by = 'Jocular' if self.created_by_jocular else 'alien'
        if self.is_master:
            logger.debug(f'{created_by} Master {self.name}')
        else:
            logger.debug(f'{created_by} Sub {self.name}')
        if not verbose:
            return

        # list essential props
        for p in default_props:
            logger.trace(f'{p} = {getattr(self, p)}')

        # some image stats
        im = self.get_image()
        logger.trace(f'min {np.min(im):.4f} max {np.max(im):.4f} mean {np.mean(im):.4f}')

        # may be other props we want to output at some point
        # logger.trace('{:9s} {:14s} ({:} days) {:5s} expo {:} filt {:} gain {:} bin {:} {:} {:}'.format(
        #     self.shape_str, 
        #     self.create_time.strftime('%d %b %y %H:%M'),
        #     self.age,
        #     self.sub_type,
        #     self.exposure,
        #     self.filter,
        #     'no' if self.gain is None else self.gain,
        #     self.binning,
        #     str(self.temperature) + 'C' if self.temperature is not None else '',
        #     str(self.nsubs) + ' subs in master' if self.nsubs is not None else ''))
        # logger.trace('from FITS: {:}'.format(fits_props))
        # logger.trace('from name: {:}'.format(name_props))



    def parse_sub(self, nm):
        # should extend this to support same parsing as calibration masters

        props = {}

        # bug here if user has 3 underscores in name....

        # pre v3 Jocular convention first
        try:
            [sub_type, filt, expo, subnum] = nm.split('_')
            props = {'sub_type': sub_type, 'filter': filt, 'exposure': int(expo[:-2]) / 1000.0}
        except:
            # parse name in one of 2 SLL formats
            nm = nm.lower()
            if nm.startswith('image_sll.'):
                nm = nm[10:]
            elif nm.startswith('image__'):
                nm = nm[7:]
            elif nm.startswith('image_'):
                nm = nm[6:]

            if nm.startswith('red') or nm.endswith('_red'):
                props['filter'] = 'R'
            elif nm.startswith('green') or nm.endswith('_green'):
                props['filter'] = 'G'
            elif nm.startswith('blue') or nm.endswith('_blue'):
                props['filter'] = 'B'
            elif nm.startswith('ha') or nm.endswith('_ha'):
                props['filter'] = 'Ha'
            elif nm[:4] in ['flat', 'dark', 'bias']:
                props['sub_type'] = nm[:4]

        return props

    def parse_master(self, nm):
        # nm  of form <type>_<key>=<value> where keys can come in any order
        # and keys are any of the things in the keymap

        prop_sep = '_'
        key_value_sep = '='
        parts = nm.split(prop_sep)
        props = {'sub_type': parts[0]}
 
        for p in parts[1:]:
            try:
                k, v = p.split(key_value_sep)
                if k.lower() in self.fitspropmap:
                    props[self.self.fitspropmap[k]] = v 
            except:
                pass

        return props

    def get_image(self):
        ''' Changed to represent subs as float32
        '''
        if self.image is None:
            try:
                with fits.open(self.path) as hdu:
                    bp = hdu[0].header['BITPIX']
                    self.image = np.array(hdu[0].data, dtype=np.float32)
                    if bp > 0:
                        self.image /= (2 ** bp)
                    self.minval = np.min(self.image) * 100
                    self.maxval = np.percentile(self.image, 99.99) * 100
                    self.meanval = np.mean(self.image) * 100
                    self.overexp = np.mean(self.image > .99) * 100   
            except Exception as e:
                logger.warning(f'cannot read image data from {self.path} ({e})')
        return self.image


    # def get_cached_image(self):
    #     ''' Images are stored in a temp folder
    #     '''




""" aligner
"""

import warnings
import numpy as np

from skimage.measure import ransac
from skimage.transform import EuclideanTransform, matrix_transform, warp

from kivy.properties import BooleanProperty, NumericProperty, StringProperty
from loguru import logger

from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.processing.starextraction import extract_stars


class Aligner(Component, JSettings):
    """RANSAC-based image aligner"""

    do_align = BooleanProperty(True)
    ideal_star_count = NumericProperty(30)
    min_stars = NumericProperty(5)
    binfac = NumericProperty(1)
    star_method = StringProperty("DoG")
    extraction_region = StringProperty('whole image')
    centroid_method = StringProperty("simple")
    do_warp = BooleanProperty(True)

    configurables = [
        (
            "do_align",
            {
                "name": "align?",
                "switch": "",
                "help": "Switching align off can help diagnose tracking issues",
            },
        ),
        (
            "extraction_region",
            {
                "name": "extract stars from",
                "options": ["whole image", "central 50%", "central 33%"],
                "help": "",
            },
        ),
        (
            "star_method",
            {
                "name": "star extraction method",
                "options": ["DoG", "photutils"],
                "help": "DoG is the original Jocular method; photutils may be faster",
            },
        ),
        (
            "centroid_method",
            {
                "name": "centroid extraction method",
                "options": ["simple", "experimental"],
                "help": "simple is the original method; experimental may be more accurate",
            },
        ),
        (
            "ideal_star_count",
            {
                "name": "ideal number of stars",
                "float": (10, 100, 5),
                "help": "Detect this many stars per sub (30 for DoG, more for photutils)",
                "fmt": "{:.0f} stars",
            },
        ),
        (
            "min_stars",
            {
                "name": "minimum number of matching stars",
                "float": (3, 20, 1),
                "help": "require this many matches for sub to be aligned",
                "fmt": "{:.0f} stars",
            },
        ),
        (
            "binfac",
            {
                "name": "bin image?",
                "float": (1, 4, 1),
                "help": "Extract stars from binned image (faster but use with care)",
                "fmt": "{:.0f}",
            },
        ),
        (
            "do_warp",
            {
                "name": "project keystars?",
                "switch": "",
                "help": "Use warp model from previous sub to shift keystars prior to matching",
            },
        ),
    ]


    def __init__(self):
        super().__init__()
        self.reset()


    def on_new_object(self):
        self.reset()


    def reset(self):
        self.keystars = None
        self.warp_model = None
        self.align_count = 0
        self.starcounts = []
        self.info("reset")


    def align(self, sub):

        if not self.do_align:
            return

        im = sub.get_image()
        w, h = im.shape

        # new
        if self.extraction_region == 'central 33%':
            ww, hh = w // 3, h // 3
            im = im[ww: w-ww, hh: h - hh]
        elif self.extraction_region == 'central 50%':
            ww, hh = w // 4, h // 4
            im = im[ww: w-ww, hh: h - hh]

        # extract stars & compute centroids
        stars = extract_stars(
            #sub.get_image(),
            im,
            star_method=self.star_method,
            centroid_method=self.centroid_method,
            binfac=self.binfac,
            target_stars=self.ideal_star_count,
            nstars=None,
            reset_threshold=self.keystars is None,
        )

        nstars = stars["nstars"]
        self.starcounts += [nstars]

        if nstars > 0:
            sub.fwhm = np.median(stars["fwhm"])

        sub.aligned = False

        # if enough stars, try to align
        if nstars > self.min_stars:
            centroids = np.transpose([stars["xcentroid"], stars["ycentroid"]])

            # first sub with stars -> keystars
            if self.keystars is None:
                self.keystars = centroids
                sub.aligned = True

            else:
                # star registration
                self.warp_model = register(
                    centroids,
                    self.keystars,
                    min_stars=self.min_stars,
                    warp_model=self.warp_model if self.do_warp else None,
                )

                if self.warp_model is not None:
                    sub.image = warp(
                        sub.image, self.warp_model, order=3, preserve_range=True
                    )
                    sub.aligned = True

        if sub.aligned:
            # default sub staus is select (set in Image)
            # but if it was not rejected by user, change it to select
            # since alignment has been successful
            if sub.status == "nalign":
                sub.status = "select"
            self.align_count += 1
        else:
            sub.status = "nalign"

        sc = np.array(self.starcounts)
        self.info(
            f"{self.align_count}/{len(sc)} subs | {np.min(sc)}-{np.max(sc)} stars"
        )


def register(stars, keystars, min_stars=None, warp_model=None):
    """Find a Euclidean transformation that matches stars
    against keystars, returning the warp model
    or None if number of inliers after RANSAC is
    less than min_stars
    """

    logger.debug("registering")

    # if we have a warp model (from prev registration),
    # use it to move keystars in the right direction
    # prior to matching stars [it does make a slight difference]
    keys = keystars.copy()
    if warp_model is not None:
        keys = matrix_transform(keys, warp_model.params)

    # find closest matches between (warped) keystars and stars
    nstars = min(len(keystars), len(stars))
    matched_stars = np.zeros((nstars, 2))
    for i, (x1, y1) in enumerate(keystars):
        if i < nstars:
            matched_stars[i, :] = stars[
                np.argmin([(x1 - x2) ** 2 + (y1 - y2) ** 2 for x2, y2 in stars])
            ]

    # do we have enough matched stars?
    if len(matched_stars) < min_stars:
        return None

    # apply RANSAC to find which matching pairs best fitting Euclidean model
    # can throw a warning in cases where no inliers (bug surely) which we ignore
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        warp_model, inliers = ransac(
            (np.array(keystars[:nstars]), matched_stars),
            EuclideanTransform,
            min_stars,
            0.5,
            max_trials=100,
        )

    # enough?
    if inliers is None or sum(inliers) < min_stars:
        return None

    return warp_model



''' Luminance channel manipulations, including B & W, stretch etc
'''

import math
import numpy as np

from skimage.transform import resize, rescale, downscale_local_mean
from skimage import filters
from skimage.morphology import disk

from astropy.stats import SigmaClip
from photutils.background import Background2D, MedianBackground

from kivy.app import App
from kivy.properties import BooleanProperty, NumericProperty, StringProperty

from jocular.settingsmanager import JSettings
from jocular.gradient import estimate_gradient, estimate_background, image_stats
from jocular.component import Component


def TNR(im, ksize=20, method='gaussian', param=1, binfac=1):
    ''' Tony's Noise Reduction
    '''
 
    ksize = int(ksize)
    if binfac > 1:
        im2 = downscale_local_mean(im, (binfac, binfac))
    else:
        im2 = im.copy()
    if method == 'gaussian':
        bkg = filters.gaussian(im2, ksize / binfac)
    elif method == 'median':
        neighbourhood = disk(radius=ksize / binfac)
        bkg = filters.rank.median(im2, neighbourhood) / 255.
    if binfac > 1:
        h, w = im.shape
        bh, bw = bkg.shape
        # use fast upsizing if result is a multiple
        if bh * binfac == h and bw * binfac == w:
            bkg = rescale(bkg, (binfac, binfac))
        else:
            bkg = resize(bkg, im.shape)

    return ((im - bkg) / (1 + np.exp(param) * np.exp(-27 * im))) + bkg


def unsharp_masking(im, radius=5, amount=1):
    return filters.unsharp_mask(im, radius=radius, amount=amount)


def fractional_bin(im, binfac=1, original_size=True):
    ''' Experimental binning with non-integer factor (binfac)
        By default, binned image is resized to original size
    '''
    if binfac < 1:
        return im

    if binfac > 3:
        binfac = 3

    binned = rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', multichannel=False)
    if original_size:
        return resize(binned, im.shape, anti_aliasing=False, mode='constant')
    else:
        return binned


def estimate_gradient_local(im):
    bkg_estimator = MedianBackground()
    sigma_clip = SigmaClip(sigma=3.)
    bkg = Background2D(im, (50, 50), filter_size=(4, 4), sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)
    return bkg.background - bkg.background_median



class Monochrome(Component, JSettings):

    tab_name = 'Luminance'

    redrawing = BooleanProperty(False)
    gradient = NumericProperty(100)
    white = NumericProperty(0.9)
    black = NumericProperty(0.1)
    p1 = NumericProperty(0.65)
    lift = NumericProperty(0)
    noise_reduction = NumericProperty(1)
    #fine = NumericProperty(0)
    autoblack = BooleanProperty(True)
    autowhite = BooleanProperty(False)
    fracbin = NumericProperty(1)
    TNR_amount = NumericProperty(0)         # was TNR_param
    TNR_kernel_size = NumericProperty(20)  # was TNR_kernel_size
    TNR_method = StringProperty('gaussian')
    unsharp_amount = NumericProperty(0)
    unsharp_radius = NumericProperty(2)
    TNR_binning = StringProperty('1')
    RL_width = NumericProperty(0)
    background_method = StringProperty('2D planar')
    imstats_units = StringProperty('percentage')

    configurables = [
        ('noise_reduction', {
            'name': 'noise reduction',
            'float': (0, 1, 0.01),
            'help': 'light touch noise reduction (0=off)'
            }),
        ('imstats_units', {
            'name': 'image stats units', 
            'options': ['percentage', 'ADUs'],
            'help': 'display min/max/mean in status line using these units'}),
        ('background_method', {
            'name': 'gradient removal', 
            'options': ['2D planar', 'regional'],
            'help': '2D planar is less flexible but more robust; regional can produce artefacts'}),
        ('TNR_method', {
            'name': 'noise reduction kernel', 
            'options': ['gaussian', 'median'],
            'help': 'convolve with this to compute background estimate'}),
        ('TNR_binning', {
            'name': 'bin image', 
            'options': ['1x1', '2x2', '3x3', '4x4', '5x5'],
            'help': 'bin image by this factor prior to background computation'}),
        ]

    save_settings = [
        'gradient', 'white', 'black', 'p1', 'lift', 'TNR_kernel_size', 
        'unsharp_radius', 'autoblack', 'autowhite'
        ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mono = None  # not 100% sure why this is needed on init
        self.view = Component.get("View")
        self.stacker = Component.get("Stacker")
        self.gui = App.get_running_app().gui


    def on_new_object(self):
        self.mono = None
        self.lum = None  # luminosity (after applying stretch, B/W etc to mono)
        self._gradient = None  # pixel-by-pixel zero-mean gradient estimate to subtract
        self._std_background = None  # std dev of background estimated from gradient
        self._blackpoint = None  # automatic estimate of blackpoint
        self._whitepoint = None  # automatic estimate of whitepoint
        self._background = None  # pixel-by-pixel background estimate used in TNR
        self._n = 0
        self.gui.set('TNR_amount', -1, update_property=True)
        self.gui.set('unsharp_amount', -.1, update_property=True)


    def on_p1(self, *args):
        self.adjust_lum()


    def on_white(self, *args):
        if not self.autowhite:
            self.adjust_lum()


    def on_black(self, *args):
        if not self.autoblack:
            self.adjust_lum()


    def on_gradient(self, *args):
        self.adjust_lum()


    def on_fracbin(self, *args):
        if self.fracbin < 1:
            return
        if self.fracbin > 3:
            return
        self.adjust_lum()


    def on_TNR_amount(self, *args):
        self.adjust_lum()


    def on_TNR_kernel_size(self, *args):
        self.adjust_lum()


    def on_unsharp_radius(self, *args):
        self.adjust_lum()


    def on_unsharp_amount(self, *args):
        self.adjust_lum()


    def on_lift(self, *args):
        self.adjust_lum()


    def on_noise_reduction(self, *args):
        self.adjust_lum()


    def on_RL_width(self, *args):
        self.adjust_lum()

    # def on_fine(self, *args):
    #     # move black by a tiny amount
    #     if not self.autoblack:
    #         newblack = min(1, max(0, self.black + 0.1 * self.fine))
    #         self.gui.set("black", float(newblack))  # causes update

    def on_autoblack(self, *args):
        if self.autoblack:
            if hasattr(self, "mono") and self.mono is not None:
                self.update_blackpoint(self.mono)
                self.adjust_lum()

    def on_autowhite(self, *args):
        if self.autowhite:
            if hasattr(self, "mono") and self.mono is not None:
                self.update_whitepoint(self.mono)
                self.adjust_lum()

    # def on_background_method(self, *args):
    #     self.update_gradient()

    def update_blackpoint(self, im):
        self._blackpoint, self._std_background = estimate_background(im)
        self.gui.set("black", float(self._blackpoint))


    def update_whitepoint(self, im):
        self._whitepoint = np.percentile(im.ravel(), 99.99)
        self.gui.set("white", float(self._whitepoint))


    def update_gradient(self, im):
        # Estimate gradient and normalise to zero mean
        if self.background_method == '2D planar':
            g = estimate_gradient(im)
            self._gradient = g - np.mean(g)
        else:
            self._gradient = estimate_gradient_local(im)


    def update_background(self, im):
        if self.TNR_method == 'gaussian':
            self._background = filters.gaussian(im, self.TNR_kernel_size)
        elif self.TNR_method == 'median':
            neighbourhood = disk(radius=self.TNR_kernel_size)
            self._background = filters.rank.median(im, neighbourhood) / 255.


    def display_sub(self, im, do_gradient=False, fwhm=None):
        ''' Called by Stacker when user selects sub, and by Capture, when 
            displaying short subs. Only compute gradient if light sub from 
            stacker, not when calibration, nor when short
            v0.5 added: also compute grad if dims changed
        '''

        # ensure we are displaying subs
        if self.stacker.viewing_stack:
            self.stacker.set_to_subs()

        # cache new image to allow user updates of B/W etc
        self.mono = im
        if do_gradient:
            self.update_gradient(im)
        # new in v0.5: if shape changes, update gradient
        elif self._gradient is not None and self._gradient.shape != im.shape:
            self.update_gradient(im) 
        if self.autoblack:
            self.update_blackpoint(im)
        if self.autowhite:
            self.update_whitepoint(im)
        self.view.display_image(self.luminance())
        self.update_info(im, fwhm=fwhm)


    def adjust_lum(self, *args):
        ''' User has changed control position so generate luminance and either
            display it (if sub or in mono mode) or advise multispectral of the update
        '''

        multispectral = Component.get("MultiSpectral")
        lum = self.luminance()
        if not self.stacker.viewing_stack:
            self.view.display_image(lum)
        else:
            if multispectral.spectral_mode == "mono":
                self.view.display_image(lum)
            else:
                multispectral.luminance_updated(lum)


    def L_changed(self, L):
        if L is not None:
            lum = self.update_lum(L)
            self.view.display_image(lum)


    def update_lum(self, im):
        ''' Create gradient-adjusted luminance image; called here by L_changed and by
            MultiSpectral. Does not directly update display.
        '''
        self.mono = im
        self.update_gradient(im)
        self.update_blackpoint(im)
        return self.luminance()


    def update_info(self, im, fwhm=None):

        stats = image_stats(im)

        satstr = '' if stats['over'] < .1 else f' sat {stats["over"]:.1f}%'
        fwhmstr = '' if fwhm is None else f'fwhm: {fwhm:.1f}px | ' 
        unitstr = 'ADU' if self.imstats_units == 'ADUs' else '%'
        rangestr = '' if self.imstats_units == 'ADUs' else 'range '

        for k, v in stats.items():
            if math.isnan(v):
                stats[k] = 0
            elif self.imstats_units == 'ADUs':
                stats[k] = int(2**16) * v
            else:
                stats[k] = 100 * v

        self.info(f"{fwhmstr}{rangestr}{stats['background']:.0f} - {stats['99.99']:.0f} {unitstr}{satstr}")


    def luminance(self, *args):
        # Applies black, white etc to current monochrome image, updating luminosity, returning the image.

        if self.mono is None:
            return None

        im = self.mono

        if self.fracbin > 1:
            im = fractional_bin(im, binfac=self.fracbin)

        # compute and display image stats
        self.update_info(im)

        # subtract some % of gradient if we have it computed (not the case for short subs)
        if (self._gradient is not None) and (self.gradient > 0.1):
            im = im - (self.gradient / 100) * self._gradient

        # set black based on automatic blackpoint estimate and lift setting
        # we also allow lift settings in non-auto case

        if self.autoblack and (self._std_background is not None):
            black = max(0, self._blackpoint - self.lift * self._std_background)

        elif self._std_background is not None:
            black = max(0, self.black - self.lift * self._std_background)

        else:
            black = self.black

        # why not working?
        if self.autowhite:
            self.update_whitepoint(im)
            white = self._whitepoint
        else:
            white = self.white

        im = (im - black) / (white - black)

        im[im > 1] = 1
        im[im < 0] = 0

        # timings: hyper:4, log: 7 asinh: 23, tanh: 6, gamma: 12
        if self._std_background is None:
            bkg = 0
        else:
            bkg = self.lift * self._std_background

        im = Component.get('Stretcher').apply_stretch(
            im,
            param=self.p1,
            NR=self.noise_reduction,
            background=bkg,
        )

        # order of these two is not clear; seem to get fewer ringing
        # artefacts if sharpening is done first

        # apply sharpening
        if self.unsharp_amount > 0 and self.unsharp_radius > 0:
            im = unsharp_masking(im, radius=self.unsharp_radius, amount=self.unsharp_amount)

        # apply noise reduction
        if self.TNR_amount > 0:
            im = TNR(
                im, 
                ksize=self.TNR_kernel_size, 
                method=self.TNR_method, 
                param=self.TNR_amount,
                binfac=int(self.TNR_binning[0]))

        # # apply sharpening
        # if self.unsharp_amount > 0 and self.unsharp_radius > 0:
        #     im = unsharp_masking(im, radius=self.unsharp_radius, amount=self.unsharp_amount)

        return im




''' Mount/scope control
	Not supported currently.
'''

from loguru import logger
from kivy.properties import StringProperty

from jocular.component import Component
from jocular.device import DeviceFamily, Device
from jocular.ascom import connect_to_ASCOM_device
from jocular.utils import toast

class Telescope(Component, DeviceFamily):

	modes = {
		'Manual': 'ManualTelescope', 
		'Simulator': 'SimulatorTelescope',
		'ASCOM': 'ASCOMTelescope'
	}
	default_mode = 'Manual'
	family = 'Telescope'

	def on_new_object(self, *args):
		logger.debug('-')
		if self.connected():
			self.device.on_new_object()

	def on_previous_object(self, *args):
		logger.debug('-')
		if self.connected():
			self.device.on_previous_object()

	def slew(self, RA=None, Dec=None):
		if self.connected():
			# get RA/Dec from DSO
			RA, Dec = Component.get('DSO').current_object_coordinates()
			if RA is not None:
				self.device.slew(RA=RA, Dec=Dec)
			else:
				toast('Cannot slew without coordinates')

	def move(self, direction=None, rate=None):
		if self.connected():
			self.device.move(direction=direction, rate=rate)

	def stop_moving(self):
		if self.connected():
			self.device.stop_moving()



class GenericTelescope(Device):
	family = StringProperty('Telescope')

	def move(self, direction=None, rate=None):
		logger.debug(f'moving {direction} at rate {rate}')

	def stop_moving(self):
		logger.debug('stop moving')


class ManualTelescope(GenericTelescope):

	def connect(self):
		self.status = 'connected'
		self.connected = True

	def slew(self, RA=None, Dec=None):
		toast('Slewing not possible for manual scope')

class SimulatorTelescope(GenericTelescope):
	pass

class ASCOMTelescope(GenericTelescope):
	# https://ascom-standards.org/Help/Developer/html/T_ASCOM_DeviceInterface_ITelescopeV3.htm

	driver = StringProperty(None)

	def disconnect(self):
		logger.debug('disconnecting AASCOM telescope')
		self.scope.Connected = False
		self.connected = False
		self.status = 'disconnected'

	def connect(self):

		if self.connected:
			return
		res = connect_to_ASCOM_device(device_type='Telescope', driver=self.driver)
		self.connected = res.get('connected', False)
		self.status = res['status']
		if self.connected:
			self.driver = res.get('driver', self.driver)
			self.scope = res['device']
			logger.info(f'current sidereal time  from scope {self.scope.SiderealTime}')
		else:
			if 'exception' in res:
				self.status += f" ({res['exception']})"



	def slew(self, RA=None, Dec=None):
		# don't forget to turn tracking on at some point...
		pass

	def move(self, direction=None, rate=None):

		logger.debug(f'moving {direction} at rate {rate}')
		''' for ASCOM, negative rate implies moving in other direction
			so need to convert directions up down etc
			We will map left right to RA axis for Eq and to azimuth for altaz
		''' 
		# map left right to RA axis for Eq and to azimuth for altaz
		axis = 0 if direction in {'left', 'right'} else 1

		# convert rate to degs/sec
		rate = {1: .1, 2: .2, 3: .3}[rate]

		# map left/down to negative rates
		rate = -rate if direction in {'left', 'down'} else rate

		# check if we can move axis
		if self.scope.CanMoveAxis(axis):
			self.scope.MoveAxis(axis, rate)

	def stop_moving(self):
		logger.debug('stop moving')
		if self.scope.CanMoveAxis(0):
			self.scope.MoveAxis(0, 0)
		if self.scope.CanMoveAxis(1):
			self.scope.MoveAxis(1, 0)






''' Handles new and previous object logic, and associated confirmation
    dialogues.
'''

import os
import shutil
from datetime import date, datetime
from loguru import logger

from kivy.app import App
from kivy.properties import BooleanProperty
from kivymd.uix.dialog import MDDialog
from kivymd.uix.button import MDFlatButton
from kivy.base import stopTouchApp

from jocular.component import Component
from jocular.utils import add_if_not_exists, generate_observation_name, move_to_dir, toast
from jocular.image import save_image, Image


class ObjectIO(Component):

    existing_object = BooleanProperty(False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.current_object_dir = None
        self.closing = False
        dformat = '%y_%m_%d'
        self.session_dir = os.path.join(self.app.get_path('captures'), date.today().strftime(dformat))
        add_if_not_exists(self.session_dir)


    def on_close(self):
        self.closing = True


    def confirm_new_object(self, *args):
        ''' user clicks new on interface, so check for save
        '''
        self.check_save(callback=self.new_object)


    def confirm_previous_object(self, *args):
        ''' user clicks previous object
        '''
        self.check_save(callback=Component.get('Observations').show_observations)


    def confirm_quit(self, *args):
        ''' user has pressed quit, so check for save
        '''
        self.check_save(callback=stopTouchApp)


    def do_callback(self):
        if hasattr(self, 'callback'):
            if self.callback is None:
                self.new_object()
            else:
                self.callback()


    def check_save(self, callback=None):
        ''' Called if user selects New, Previous or Quit
        '''

        logger.debug('')
        self.callback = callback

        # nothing to save
        if not Component.any_changes():
            self.do_callback()
            return

        # something has changed so confirm with user
        changes = Component.changes()

        self.dialog = MDDialog(
            auto_dismiss=False, 
            text=f"Save current observation?\n\nReason(s): {', '.join(list(changes.values()))}", 
            buttons=[
                MDFlatButton(text="SAVE", on_press=self._save),
                MDFlatButton(text="NO",  on_press=self._no_save),
                MDFlatButton(text="CANCEL", on_press=self._cancel)
            ])
        self.dialog.open()


    def _save(self, *args):
        self.dialog.dismiss()
        self.save()


    def _no_save(self, callback, *args):
        self.dialog.dismiss()
        # for any live observations, move to delete directory
        if not self.existing_object:
            self.delete_file(self.current_object_dir)
            toast('moved FITs for this object to delete directory', 3)
        self.do_callback()


    def _cancel(self, *args):
        self.dialog.dismiss()


    def save(self, *args):
        ''' save sub metadata and potentially master also
        '''

        if self.current_object_dir is None:
            logger.warning('trying to save but current_object_dir is None')
            return

        stacker = Component.get('Stacker')
        metadata = Component.get('Metadata')

        name = Component.get('DSO').Name
        sub_type = stacker.get_prop('sub_type')
        calib = sub_type in {'dark', 'flat', 'bias'}

        exposure = stacker.get_prop('exposure')

        # try to get temperature from sensor; if not from user
        temperature = Component.get('Camera').get_sensor_temperature()
        if temperature is None:
            temperature = Component.get('Session').temperature

        if calib:
            if temperature == -40:
                temperature = None
            capture_props = {
                'exposure': exposure, 
                'temperature': temperature,
                'filter': stacker.get_prop('filter'),
                'gain': stacker.get_prop('gain'),
                'offset': stacker.get_prop('offset'),
                'binning': stacker.get_prop('binning'),
                'camera': stacker.get_prop('camera'),
                'calibration_method': stacker.get_prop('calibration_method'),
                'sub_type': sub_type
            }
            Component.get('Calibrator').create_master(capture_props=capture_props)

        metadata.set(
            {'exposure': exposure, 
            'sub_type': sub_type,
            'temperature': temperature})

        oldpath = self.current_object_dir
        Component.save_object()

        # check if name has been changed
        name = metadata.get('Name', default='')
        if name:
            session_dir, object_dir = os.path.split(self.current_object_dir)
            if name != object_dir:
                # user has changed name, so generate a (unique) new folder name
                new_name = generate_observation_name(session_dir, prefix=name)
                # change directory name; any problem, don't bother
                try:
                    os.rename(self.current_object_dir, 
                        os.path.join(session_dir, new_name)) 
                    self.current_object_dir = os.path.join(session_dir, new_name)
                except Exception as e:
                    logger.exception(f'cannot change name ({e})')

        # save metadata, and if successful update observations
        newpath = self.current_object_dir
        try:
            ''' kludge for now: temperature at least may have been changed by session 
                during save_object so ensure metadata is reset with these new values
            '''
            metadata.set(
                {'exposure': exposure, 
                'sub_type': sub_type,
                'temperature': temperature})
            # metadata.save(newpath, change_fits_headers=content.change_fits_headers)
            metadata.save(newpath)
            Component.get('Observations').update(oldpath, newpath)
            Component.get('ObservingList').new_observation()
            toast(f'Saved to {newpath}')
        except Exception as e:
            logger.exception(f'OSError saving info3.json to {newpath} ({e})')

        # do next action
        self.do_callback()


    def new_object(self, *args):
        if self.closing:
            return
        self.current_object_dir = None
        self.existing_object = False
        self.sub_type = None
        Component.initialise_new_object()
        self.app.gui.set('frame_script', True, update_property=True)


    def load_previous(self, path):
        # previous will have been saved by this point
        self.existing_object = True
        gui = self.app.gui
        gui.enable()
        gui.set('show_reticle', False, update_property=True)
        self.current_object_dir = path
        Component.initialise_previous_object()


    def new_sub(self, data=None, name=None, capture_props=None):
        ''' Called by Capture or WatchedCamera
            capture_props is a dictionary of properties such
            as exposure, sub type, gain etc
        '''

        if data is None or name is None:
            return

        stacker = Component.get('Stacker')

        # check dimensions
        if not stacker.is_empty():
            if data.shape != stacker.subs[0].image.shape:
                toast(f'sub dimensions {data.shape} incompatible with current stack {stacker.subs[0].image.shape}')
                return

        logger.debug(f'New sub | {capture_props}')
        sub_type = capture_props['sub_type']
 
        if self.current_object_dir is None:
            # new object, so check if calibration or light
            if sub_type in {'dark', 'flat', 'bias'}:
                self.current_object_dir = os.path.join(self.session_dir, 
                    generate_observation_name(self.session_dir, prefix=sub_type))
                # not clear if this next line is needed bcs stacker doesn't
                # have a sub_type component
                stacker.sub_type = sub_type 
                Component.get('DSO').Name = sub_type
            else:
                self.current_object_dir = os.path.join(self.session_dir, 
                    generate_observation_name(self.session_dir, prefix=Component.get('DSO').Name))
            add_if_not_exists(self.current_object_dir)

        path = os.path.join(self.current_object_dir, name)

        try:
            save_image(data=data, path=path, capture_props=capture_props)
            stacker.add_sub(Image(path))
        except Exception as e:
            logger.exception(f'cannot add sub to stack ({e})')
           

    def save_original(self, path):
        ''' Move non-Jocular sub from path (in Watched) to current object subdirectory
        '''
        if self.current_object_dir is None:
            self.current_object_dir = os.path.join(self.session_dir, 
                generate_observation_name(self.session_dir, prefix=Component.get('DSO').Name))
            add_if_not_exists(self.current_object_dir)
        move_to_dir(path, os.path.join(self.current_object_dir, 'originals'))


    def delete_file(self, path):
        ''' Move file to delete directory under today's date for convenience
        '''
        try:
            delete_dir = self.app.get_path('deleted')
            today = datetime.now().strftime('%d_%b_%y')
            delete_path = os.path.join(delete_dir, today)
            if not os.path.exists(delete_path):
                os.mkdir(delete_path)
            dname = os.path.join(delete_path,
                f"{os.path.basename(path)}_{datetime.now().strftime('%H_%M_%S.%f')}")
            shutil.move(path, dname)
        except Exception as e:
            logger.exception(f'Problem deleting {path} ({e})')




''' Support for RA/Dec text representations and conversions.
'''
import math
from loguru import logger

class RA(float):
    def __new__(cls, ra):
        if type(ra) == str:
            try:
                h, m ,s = ra.split()
                if h[-1] == 'h':
                    h = h[:-1]
                ra = float(h)*15 + float(m)/4 + float(s)/240
                if (ra < 0) | (ra > 360):
                    return float('nan')
            except:
                return float('nan')
        return float.__new__(cls, ra)
 
    def __str__(self):
        if math.isnan(self):
            return ''
        hr, rest = divmod(self, 15)
        minit, rest = divmod(rest*4, 1)
        return f'{int(hr):02d}h {int(minit):02d} {rest*60:02.0f}'

    @classmethod
    def parse(cls, s):
        ''' We allow forms such as 23h34 21, 23 34 21, 23h34'21"
        '''
        try:
            s = s.replace('h', ' ').replace('"',' ').replace("'", ' ')
            parts = s.split()
            if len(parts) == 3:
                hrs, mins, secs = parts
                hrs = int(hrs)
                mins = int(mins)
                secs = float(secs)
                if (hrs >= 0) & (hrs < 24) & (mins >=0) & (mins < 60) & (secs >= 0) & (secs < 60):
                    return float(hrs*15 + mins/4 + secs/240 )
            return None
        except Exception as e:
            logger.warning(f'Problem parsing RA {s} ({e})')
            return None


class Dec(float):
    def __new__(cls, dec):
        if type(dec) == str:
            try:
                d, m, s = dec.split()
                if d[-1] == '\u00b0':
                    d = d[:-1]
                if d[0] == '-':
                    dec = float(d) - float(m)/60 - float(s)/3600
                else:
                    dec = float(d) + float(m)/60 + float(s)/3600
                if (dec < -90) | (dec > 90):
                    return float('nan')
            except:
                return float('nan')
        return float.__new__(cls, dec)

    def __str__(self):
        if math.isnan(self):
            return ''
        if self >= 0:
            dd, rest = divmod(self, 1)
            dmin, rest = divmod(rest*60, 1)
            return f"+{int(dd):d}\u00b0 {int(dmin):02d} {rest*60:02.0f}"
        else:
            dd, rest = divmod(-self, 1)
            dmin, rest = divmod(rest*60, 1)
            return f"-{int(dd):d}\u00b0 {int(dmin):02d} {rest*60:02.0f}"

    @classmethod
    def parse(cls, s):
        ''' We allow forms such as 23d34 21, 23 34 21, 23<degree symbol>34'21.3"
        '''
        try:
            s = s.replace('d', ' ').replace('"',' ').replace("'", ' ').replace('\u00b0', ' ')
            parts = s.split()
            if len(parts) == 3:
                degs, mins, secs = parts
                degs = int(degs)
                mins = int(mins)
                secs = float(secs)
                if (degs >= -90) & (degs < 90) & (mins >=0) & (mins < 60) & (secs >= 0) & (secs <= 60):
                    if degs >= 0:
                        return float(degs + mins/60 + secs/3600)
                    return float(degs - mins/60 - secs/3600)
            return None
        except Exception as e:
            logger.warning(f'Problem parsing Dec {s} ({e})')
            return None




''' DeviceManager: a Component that manages different device families 
	e.g. Telescope, Camera, FilterWheel via a GUI element that permits 
	selection/connection/disconnection
'''


from functools import partial

from kivy.app import App
from loguru import logger
from kivy.metrics import dp
from kivy.uix.spinner import Spinner
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout

from jocular.component import Component
from jocular.widgets.widgets import LabelL
from jocular.widgets.icons import get_icon
from jocular.formwidgets import configurable_to_widget
from jocular.panel import Panel


class DeviceManager(Component, Panel):


	def __init__(self, **args):
		super().__init__(**args)
		self.app = App.get_running_app()
		self.status = {}
		self.instances = {}
		self.current_panel = 'Camera'
		self.connect_buttons = {}
		self.connect_dots = {}
		self.app.gui.add_widget(self)


	def register(self, device, name=None):
		''' keep a record of all device instances
		'''
		self.instances[name] = device
		logger.debug(f'registered device for {name}')


	def on_leave(self, *args):
		self.current_device.apply_and_save_settings()
		self.hide()


	def on_show(self):
		''' Main device manager panel that handles mode selection and connection,
			and links to configuration of current devices. 
		'''

		self.contents.clear_widgets()
		self.header.clear_widgets()
		self.contents.width = dp(800) # was 600

		# top spinner
		hb = BoxLayout(size_hint=(1, None), height=dp(32))
		hb.add_widget(Label(size_hint=(1, 1)))
		self.spinner = Spinner(
			text=self.current_panel,
			values=self.instances.keys(),
			size_hint=(None, 1), width=dp(130), font_size='20sp')
		self.spinner.bind(text=self.device_panel_changed)
		hb.add_widget(self.spinner)
		hb.add_widget(Label(size_hint=(1, 1)))
		self.header.add_widget(hb)

		# spacer
		self.contents.add_widget(Label(size_hint=(1, None), height=dp(20)))

		# panel for device-specific contents will go
		self.panel = BoxLayout(
			orientation='vertical', 
			size_hint=(1, 1))

		self.contents.add_widget(self.panel)

		self.show_panel()


	def device_panel_changed(self, spinner, *args):
		self.current_panel = spinner.text
		self.show_panel()


	def show_panel(self):
		''' update panel with current class settings
		'''

		self.panel.clear_widgets()

		device = self.current_panel
		current_device = Component.get(device).device

		# connection line
		bh = BoxLayout(
			size_hint=(1, None), 
			height=dp(32),
			spacing=dp(20))

		bh.add_widget(Label(size_hint=(1, 1)))

		# connection status
		lab = self.connect_dots[device] = LabelL(
			size_hint=(None, 1), 
			width=dp(24), 
			markup=True,
			text=get_icon(
				'dot', 
				font_size=12,
				color='g' if current_device.connected else 'r'))
		bh.add_widget(lab)

		# device chooser
		spinner = Spinner(size_hint=(None, 1), width=dp(120),
			text=Component.get(device).settings['current_mode'],
			values=Component.get(device).modes.keys())
		spinner.bind(text=partial(self.mode_changed, device))
		bh.add_widget(spinner)

		# connect/disconnect button
		but = self.connect_buttons[device] = Button(size_hint=(None, 1), width=dp(120),
			text='disconnect...' if current_device.connected else 'connect...', 
			on_press=partial(self.connect, device)) 
		bh.add_widget(but)
		
		bh.add_widget(Label(size_hint=(1, 1)))

		self.panel.add_widget(bh)

		# connection status message
		bh = BoxLayout(padding=(10, 1), size_hint=(1, None), height=dp(40))
		status = self.status[device] = Label(text=current_device.status, 
			size_hint=(1, 1), color=(.5, .5, .5, 1))
		bh.add_widget(status)
		self.panel.add_widget(bh)

		# config panel: device-specific config params
		self.config_panel = BoxLayout(
			orientation='vertical', 
			size_hint=(1, 1))

		self.panel.add_widget(self.config_panel)

		self.config(device)


	def mode_changed(self, device, spinner, mode):
		Component.get(device).set_mode(mode)
		self.config(device)


	def connect(self, device, widget=None):
		try:
			if self.connect_buttons[device].text == 'connect...':
				Component.get(device).connect()
			else:
				Component.get(device).disconnect()
			Component.get(device).save()
		except Exception as e:
			logger.exception(e)


	def status_changed(self, device, status):
		if device in self.status:
			self.status[device].text = status


	def connection_changed(self, device, connected):
		if device in self.connect_dots:
			self.connect_dots[device].text = get_icon(
				'dot', color=('g' if connected else 'r'))
			Component.get(device).info('not connected')
		if device in self.connect_buttons:
			self.connect_buttons[device].text = 'disconnect...' if connected else 'connect...'
			Component.get(device).info('connected')


	def config(self, device, *args):
		''' user wants to configure device
		'''
		logger.debug(f'Configuring {device} device')
		try:
			self.current_device = Component.get(device).device
			self.changed_settings = {}
			if self.current_device is not None:
				self.show_device_config_panel(name=device, device=self.current_device)
		except Exception as e:
			logger.exception(e)


	def show_device_config_panel(self, name=None, device=None):
		''' Build device config panel 
		'''
		self.config_panel.clear_widgets()
		self.config_panel.add_widget(Label(size_hint=(1, 1))) # spacer
		for pname, pspec in device.configurables:
			self.config_panel.add_widget(configurable_to_widget(
				text=pspec.get('name', pname),
				name=pname,
				spec=pspec,
				helptext=pspec.get('help', ''),
				initval=getattr(self.current_device, pname), 
				changed=device.setting_changed))
		self.config_panel.add_widget(Label(size_hint=(1, 1))) # spacer


	def on_touch_down(self, touch):
		handled = super().on_touch_down(touch)
		if self.collide_point(*touch.pos):
			return True
		return handled



''' Serves stars and DSO catalogues to other components, notably the
    platesolver, annotator, and observing list
'''

import os
import glob
import csv
import json
import math
from copy import deepcopy
import numpy as np
from kivy.app import App
from loguru import logger
from kivy.properties import BooleanProperty

from jocular.component import Component
from jocular.settingsmanager import JSettings


def intstep(d, step):
    return int(np.floor(step * np.floor(d / step)))


class Catalogues(Component, JSettings):

    VS = BooleanProperty(True)
    WDS = BooleanProperty(True)
    SkiffSpectralClasses = BooleanProperty(True)
    milliquas = BooleanProperty(True)
    Hyperleda = BooleanProperty(True)

    tab_name = 'Catalogues'
    configurables = [
        ('VS', {'name': 'annotate variable stars?', 'switch': '',
            'help': 'Switch this off if you end up with crowded annotations'}),
        ('WDS', {'name': 'annotate double stars?', 'switch': ''}),
        ('SkiffSpectralClasses', {'name': 'annotate spectra classes?', 'switch': ''}),
        ('milliquas', {'name': 'annotate quasars?', 'switch': ''}),
        ('Hyperleda', {'name': 'annotate faint galaxies?', 'switch': ''})
        ]

    props = ['Name', 'RA', 'Dec', 'Con', 'OT', 'Mag', 'Diam', 'Other']


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.star_db = None

        ''' load object types and normalise colour values to 0-1 
        '''
        with open(self.app.get_path('object_types.json'), 'r') as f:
            self.object_types = json.load(f)
        for v in self.object_types.values():
            v['col'] = [i/255. for i in v['col']]

        self.dsos = None
        self.on_new_object()


    def get_object_types(self):
        return self.object_types


    def on_new_object(self):
        ''' Every time we get a new object, check for the existence
            of the catalogues. This allows a user to load them without
            restarting Jocular
        '''
        if self.star_db is None:
            path = self.app.get_path('star_db')
            if os.path.exists(path):
                try:
                    self.star_db  = np.load(path)
                    logger.info('loaded platesolving star database')
                except Exception as e:
                    logger.warning(f'cannot load star database from {path}: {e}')


    def has_platesolving_db(self):
        return self.star_db is not None


    def get_star_db(self):
        return self.star_db


    def in_tile(self, tile, ras, decs):
        ''' return indices of coordinates within tile
        '''
        rmin, rmax = tile['min_ra'], tile['max_ra']
        dmin, dmax = tile['min_dec'], tile['max_dec']
        if rmin < rmax:
            return (decs >= dmin) & (decs < dmax) & (ras >= rmin) & (ras < rmax)
        else:
            return (decs >= dmin) & (decs < dmax) & ((ras >= rmin) | (ras < rmax))


    def get_tiles(self, tile, rstep, dstep):
        ''' return database tiles covering region defined in tile
        '''
        rmin, rmax = tile['min_ra'], tile['max_ra']
        dmin, dmax = tile['min_dec'], tile['max_dec']

        dec_tiles = np.arange(intstep(dmin, dstep), intstep(dmax, dstep) + dstep, dstep)
        if rmin < rmax:
            ra_tiles = np.arange(intstep(rmin, rstep), intstep(rmax, rstep) + rstep, rstep)
        else:
            # handle wraparound
            ra_tiles = \
                list(range(intstep(rmin, rstep), 360, rstep)) + \
                list(range(0, intstep(rmax, rstep) + rstep, rstep))
        return ra_tiles, dec_tiles        


    def get_platesolving_stars(self, tile=None):
        # get stars in tile by reading npys that are spaced at 30 x 10 degrees in RA/Dec

        if self.star_db is None or tile is None:
            return

        ra_tiles, dec_tiles = self.get_tiles(tile, 30, 10)
        ras, decs, mags = np.array([]), np.array([]), np.array([])
        for ra in ra_tiles:
            for dec in dec_tiles:
                quad = f'r{ra}_d{dec}_'
                ras = np.append(ras, self.star_db[quad+'ra'])
                decs = np.append(decs, self.star_db[quad+'dec'])
                mags = np.append(mags, self.star_db[quad+'mag']/100.) # mags are ints * 100

        locs = self.in_tile(tile, ras, decs)
        return ras[locs], decs[locs], mags[locs]


    def get_basic_dsos(self):
        if self.dsos is None:
            self.load_basic_dsos()
        return self.dsos


    def load_basic_dsos(self):
        ''' load shipped catalogues, user catalogues, and user objects (i.e. those
            resulting from user observations)
        '''

        shipped = glob.glob(os.path.join(self.app.get_path('dsos'), '*.csv'))
        usercats = glob.glob(os.path.join(self.app.get_path('catalogues'), '*.csv'))
        allcats = [c for c in shipped + usercats if not c.endswith('user_objects.csv')] # legacy

        self.dsos = {}
        for cat in allcats:
            cnt = 0
            try:
                with open(cat, newline='') as f:
                    reader = csv.DictReader(f)
                    for d in reader:
                        d['Name'] = d['Name'].strip()
                        nm = f"{d['Name']}/{d['OT']}".upper()
                        self.dsos[nm] = d
                        self.dsos[nm]['Usr'] = ''
                        cnt += 1
                logger.info(f'loaded {cnt} DSOs from {cat}')
            except Exception as e:
                logger.error(f'problem loading/reading catalogue {cat} ({e})')

        # read user objects or initialise
        try:
            with open(os.path.join(self.app.get_path('catalogues'), 'user_objects.json')) as f:
                user_objects = json.load(f)
            # add to dsos only if not already present, and mark as user objects
            for k, v in user_objects.items():
                ku = k.upper()
                if ku not in self.dsos:
                    self.dsos[ku] = v
                    self.dsos[ku]['Usr'] = 'Y'
        except Exception as e:
            logger.error(f'problem loading/reading user_objects.json ({e})')

        # update user objects (test)
        self.save_user_objects()

        # ensure all fields are present and consistent
        for v in self.dsos.values():
            for col in ['RA', 'Dec', 'Mag', 'Diam']:
                val = v.get(col, '')
                try:
                    v[col] = float(val)
                except:
                    v[col] = math.nan
            for col in ['OT', 'Con']:
                v[col] = v.get(col, '').upper()
            v['Other'] = v.get('Other', '')


    def lookup(self, name, OT=None):
        ''' lookup object details for name
            If name is in form name/OT then return match or None
            If name is in form name (no OT) then return data for all matches
        '''

        name = name.upper()

        if OT is not None and OT and '/' not in name:
            name = f'{name}/{OT}'

        logger.debug(name)
        dsos = self.get_basic_dsos()
        if name in dsos:
            return dsos[name]

        # if just provided with a name and no OT, check if there is a unique match
        if '/' not in name:
            name += '/'
            matches = [n for n in dsos if n.startswith(name)]
            if len(matches) == 1:
                return dsos[matches[0]]

        return None


    def update_user_object(self, props):
        ''' Called from DSO when object is saved and is new or altered
        '''

        name = f"{props['Name']}/{props['OT']}".upper()

        if name in self.dsos and self.dsos[name]['Usr'] != 'Y':
            logger.warning(f'cannot update non-user-defined DSO {name}')
            return

        # update main DSO list
        self.dsos[name] = props.copy()
        self.dsos[name]['Usr'] = 'Y'

        self.save_user_objects()


    def save_user_objects(self):
        fields = ['Name', 'OT', 'Con', 'RA', 'Dec', 'Diam', 'Mag', 'Other']
        udict = {}
        for k, v in self.dsos.items():
            if v['Usr'] == 'Y':
                udict[k] = {k1: v1 for k1, v1 in v.items() if k1 in fields}
        try:
            with open(os.path.join(self.app.get_path('catalogues'), 'user_objects.json'), 'w') as f:
                json.dump(udict, f, indent=1)
        except Exception as e:
            logger.warning(f'Unable to save user_objects.csv ({e})')


    def delete_user_object(self, name):
        ''' Remove user object and save user objects list
        '''

        name = name.upper()
        if name in self.dsos and self.dsos[name]['Usr'] == 'Y':
            del self.dsos[name]
            self.save_user_objects()
       

    def is_excluded(self, catname):
        return hasattr(self, catname) and not getattr(self, catname)


    def get_annotation_dsos(self, tile=None):
        ''' Retrieve all DSOs in the specified tile. Use basic DSOs combined
            with any deeper DSOs available. Return as a dict with dso names as
            keys and values are dicts of object properties

            To do: streamline using in_tile etc so RA wraparound is handled in one place
        '''

        if tile is None:
            return {}

        dsos = self.get_basic_dsos()
        dmin, dmax = tile['min_dec'], tile['max_dec']
        in_dec = {k: v for k, v in dsos.items() 
            if (v['Dec'] >= dmin) & (v['Dec'] < dmax)}
        rmin, rmax = tile['min_ra'], tile['max_ra']
        if rmin < rmax:
            matches = {k: v for k, v in in_dec.items() 
                if (v['RA'] >= rmin) & (v['RA'] < rmax)}
        else:
            # handle RA wrap-around
            matches = {k: v for k, v in in_dec.items() 
                if (v['RA'] >= rmin) | (v['RA'] < rmax)}

        # load any deep catalogues that user has placed in catalogues
        deepcats = glob.glob(os.path.join(self.app.get_path('catalogues'), '*.npz'))
        for cat in deepcats:
            if not self.is_excluded(os.path.basename(cat)[:-4]):
                matches = {**matches, **self.load_deepcat(cat, tile)}

        # return copy in case receiver decides to modify things
        return deepcopy(matches)


    def load_deepcat(self, nm, tile):

        cat = np.load(nm, allow_pickle=True)
        try:
            rstep, dstep = cat['ra_step'][0], cat['dec_step'][0]
        except:
            return {}

        ra_tiles, dec_tiles = self.get_tiles(tile, rstep, dstep)

        # extract objects from all necessary tiles
        cols = cat['columns']
        OT = cat['OT'][0]
        catname = cat['catname'][0]

        matches = {}
        for ra in ra_tiles:
            for dec in dec_tiles:
                subdat = cat[f'r{ra}_d{dec}'][0]
                ras, decs = subdat['RA'], subdat['Dec']
                locs = self.in_tile(tile, ras, decs)
                for l in np.where(locs)[0]:
                    matches[subdat['Name'][l] + '/' + OT] = {c: subdat[c][l] for c in cols}

        # add object type and catalogue
        for k, v in matches.items():
            v['OT'] = OT
            v['Cat'] = catname

        return matches




''' Manages calibration, alignment and stacking of incoming subs
'''

import os
import glob
import math
import numpy as np
from datetime import datetime
from loguru import logger
from astropy.io import fits

from kivy.app import App
from kivy.properties import (BooleanProperty, NumericProperty, ListProperty)
from kivy.clock import Clock
from kivymd.uix.button import MDFlatButton
from kivymd.uix.dialog import MDDialog
from kivy.core.window import Window
from kivy.metrics import dp

from jocular.table import Table
from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.utils import s_to_minsec, move_to_dir, purify_name, toast
from jocular.image import Image, fits_in_dir
from jocular.widgets.widgets import JSlider

from kivy.lang import Builder

Builder.load_string(
    '''

<MySpinnerOption@SpinnerOption>:
    size_hint_y: None
    height: "24dp"
    background_color: .1, .1, .1, 1

''')


date_time_format = '%d %b %y %H:%M:%S'


def handleNAN(x):
    if type(x) == str:
        return x
    if x is None:
        return math.nan
    return math.nan if math.isnan(x) else x



class Stacker(Component, JSettings):

    save_settings = ['speed']

    selected_sub = NumericProperty(-1) # index of currently selected sub starts at 0
    viewing_stack = BooleanProperty(False)
    subs = ListProperty([])
    animating = BooleanProperty(False)
    speed = NumericProperty(2)
    confirm_before_deleting_stack = BooleanProperty(True)
    reload_rejected = BooleanProperty(False)

    configurables = [
        ('confirm_before_deleting_stack', {
            'name': 'confirm before clearing stack?', 
            'switch': '',
            'help': 'Clearing deletes all the current subs in the stack'}),
        ('reload_rejected', {'name': 'reload rejected subs?', 
            'switch': '',
            'help': 'include subs that were previously marked as rejected'})
        ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        # self.on_spectral_mode()
        self.sub_colors = {'select': (0, 1, 0, .6), 
            'reject': (.5, .5, .5, .6), 
            'nalign': (1, 0, 0, .6),
            'delete': (.1, .1, .1, .6)}
        # initialise speed from settings
        config = self.app.gui.config
        self.speed = config.get('speed', 2)


    def on_new_object(self):
        self.reset()
        # self.app.gui.enable(['clear_stack'])


    def on_previous_object(self):

        # do the standard resets
        self.reset()

        # don't allow users to clear stack for previous objects
        # self.app.gui.disable(['clear_stack'])

        # and all the specific stuff to reload previous
        cod = Component.get('ObjectIO').current_object_dir
        settings = Component.get('Metadata').get({'sub_type', 'rejected', 'exposure'})
        rejected = set(settings.get('rejected', []))
        exposure = settings.get('exposure', None)
        sub_type = settings.get('sub_type', 'light')
        if self.reload_rejected:
            for f in fits_in_dir(os.path.join(cod, 'rejects')):
                try:
                    bn = os.path.basename(f)
                    os.rename(f, os.path.join(cod, bn))
                    rejected.add(bn)
                except Exception as e:
                    logger.exception(f'problem moving rejected on load {f} ({e})')

        # now load all (which now will include prev rejects)
        sub_num = 0
        for f in fits_in_dir(cod):
            try:
                im = Image(f)
                im.status = 'reject' if os.path.basename(f) in rejected else 'select'
                if im.exposure is None:
                    im.exposure = exposure
                if im.sub_type is None:
                    im.sub_type=sub_type
                sub_num += 1
                im.stack_num = sub_num
                self.subs.append(im)
            except Exception as e:
                logger.exception(f'problem reading image {f} ({e})')

        # if not self.is_empty():
        #      self.original_exposure = self.subs[0].exposure

        # store rejected to see if anything has changed and needs saving
        self.orig_rejects = {os.path.splitext(r)[0].lower() for r in rejected}

        if self.is_empty():
            return

        self.recompute()

        # get filter/exposure/subtype and send to CaptureScript
        expo = self.get_prop('exposure', average=True)

        if expo is None:
            expo = Component.get('Metadata').get('exposure')

        Component.get('CaptureScript').set_external_details(
            exposure=0 if expo is None else expo, 
            sub_type=self.get_prop('sub_type'), 
            filt=''.join({s.filter for s in self.subs}))


    def reset(self):
        # Called when we have a new object and when user clears stack
        self.subs_table_records = None
        self.stop_load() # new in v0.5
        # self.stack_cache = {}
        self.orig_rejects = set({}) # new
        self.subs.clear()
        self.selected_sub = -1
        self.update_stack_scroller()
        # might add option to allow default to show stack after reset
        # in which case it would be this line instead of the next
        # self.app.gui.set('sub_or_stack', 'stack', update_property=True)
        self.set_to_subs()
        Component.get('View').reset()  # might not be needed as View also does a reset!
        self.info('reset')


    def on_save_object(self):
        # move rejected subs to 'rejected'; we won't touch non-aligned as they are salvageable often
        cod = Component.get('ObjectIO').current_object_dir
        for s in self.subs:
            if s.status == 'reject':
                move_to_dir(os.path.join(cod, s.fullname), 'rejects')


    def is_empty(self):
        return len(self.subs) == 0


    def update_status(self):
        d = self.describe()
        if d:
            self.info('{:} | {:}x{:} | {:}'.format(
                s_to_minsec(d['total_exposure']), 
                d['nsubs'], 
                s_to_minsec(d['sub_exposure']), 
                d['filters']))


    def describe(self):
        # return info useful to snapshotter and status update

        # viewing single sub
        if len(self.subs) > 0 and not self.viewing_stack:
            s = self.subs[self.selected_sub]
            return {
                'nsubs': 1,
                'total_exposure': s.exposure if s.exposure else 0,
                'sub_exposure': s.exposure if s.exposure else 0,
                'multiple_exposures': False,
                'gain': s.gain if s.gain else 0,
                'dark': 'dark' in s.calibrations,
                'flat': 'flat' in s.calibrations,
                'bias': 'bias' in s.calibrations,
                'filters': f'filter: {s.filter}'}

        # remove any not selected
        subs = [s for s in self.subs[:self.selected_sub+1] if s.status == 'select']
        if len(subs) == 0:
            return None

        # check with multispectral which filters are actually being displayed
        filters_being_displayed = Component.get('MultiSpectral').filters_being_displayed()

        # restrict to these filters
        subs = [s for s in subs if s.filter in filters_being_displayed]

        expos = [s.exposure if s.exposure else 0 for s in subs]
        filts = [s.filter for s in subs]
        fstr = ''

        gains = [s.gain if s.gain else 0 for s in subs]

        # ftypes = Component.get('FilterChooser').get_filter_types()
        ftypes = set(filts)

        for f in ftypes:
            if filts.count(f) > 0:
                fstr += f'{filts.count(f)}{f} '

        return {
            'nsubs': len(subs),
            'total_exposure': sum(expos),
            'gain': np.mean(gains),
            'sub_exposure': max(set(expos), key=expos.count),
            'multiple_exposures': len(set(expos)) > 1,
            'dark': 'dark' in subs[0].calibrations,
            'flat': 'flat' in subs[0].calibrations,
            'bias': 'bias' in subs[0].calibrations,
            'filters': fstr.strip()}


    def get_filters(self):
        if self.is_empty():
            return None
        return {s.filter for s in self.subs[:self.selected_sub+1] if s.status == 'select'}


    def delete_sub_or_stack(self, *args):
        ''' called when user clicks 'trash' icon
        '''
        if self.is_empty():
            return
        #if self.sub_or_stack == 'stack':
        if self.viewing_stack:
            self.delete_stack()
        else:
            self.delete_sub()


    def delete_stack(self, *args):
        ''' Clear existing stack by moving/deleting FITs and resetting
        '''
        if Component.get('ObjectIO').existing_object:
            toast('Cannot delete stack for previous capture')

        elif self.confirm_before_deleting_stack:
            self.dialog = MDDialog(
                auto_dismiss=False,
                text="Are you sure you wish to clear the stack?",
                buttons=[
                    MDFlatButton(text="YES", 
                        text_color=self.app.theme_cls.primary_color,
                        on_press=self._delete_stack),
                    MDFlatButton(
                        text="CANCEL", 
                        text_color=self.app.theme_cls.primary_color,
                        on_press=self._cancel)
                ],
            )
            self.dialog.open()
        else:
            self._delete_stack()


    def _cancel(self, *args):
        self.dialog.dismiss()


    def _delete_stack(self, *args):
        ''' clear stack auxilliary method
        '''
        if hasattr(self, 'dialog'):
            self.dialog.dismiss()
        objio = Component.get('ObjectIO')
        cod = objio.current_object_dir
        for nm in glob.glob(os.path.join(cod, '*.fit*')):
            objio.delete_file(os.path.join(cod, nm))

        # perform necessary resets
        Component.get('Aligner').reset()
        self.reset()
        Component.get('Capture').reset(stop_capturing=False)
        self.check_for_change()


    def delete_sub(self):
        ''' delete current sub, except for first (at present)
        '''
        sub_num = self.selected_sub
        if sub_num == 0:
            toast('Cannot delete first sub')
            return
        objio = Component.get('ObjectIO')
        cod = objio.current_object_dir
        objio.delete_file(os.path.join(cod, self.subs[sub_num].name + '.fit'))        
        del self.subs[sub_num]
        self.selected_sub -= 1
        self.check_for_change()


    def on_animating(self, *args):
        ''' animate subs or stack
        '''
        if self.is_empty():
            return
        if self.animating:
            self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)
            # construct slider if we haven't already
            if not hasattr(self, 'speed_slider'):
                self.speed_slider = JSlider(
                    size_hint=(None, None), width=dp(200), height=dp(30), 
                    step=0.2, min=0.3, max=7, value=self.speed)
                self.speed_slider.bind(value=self.speed_changed)
                self.app.gui.add_widget(self.speed_slider)
            self.speed_slider.pos_hint = {'center_x': .5, 'center_y': .90} 

        else:
            # hide scroller and cancel play
            self.speed_slider.pos_hint = {'center_x': 10, 'center_y': .90} 
            if hasattr(self, 'play_event'):
                Clock.unschedule(self.play_event)


    def speed_changed(self, widget, *args):
        self.speed = widget.value


    def on_viewing_stack(self, *args):
        self.on_selected_sub()


    def on_speed(self, *args):
        if self.animating:
            Clock.unschedule(self.play_event)
            self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)


    def first_sub(self, *args):
        self.selected_sub = 0


    def last_sub(self, *args):
        self.selected_sub = len(self.subs) - 1


    def previous_sub(self, *args):
        if self.selected_sub > 0:
            self.selected_sub -= 1


    def next_sub(self, *args):
        if self.selected_sub < len(self.subs) - 1:
            self.selected_sub += 1


    def set_to_subs(self):
        # self.app.gui.set('sub_or_stack', 'sub', update_property=True)
        self.app.gui.set('viewing_stack', False, update_property=True)


    def toggle_selected_sub(self, *args):
        # user selects displayed sub icon to toggle its selected status

        if self.is_empty():
            return

        current_sub = self.selected_sub
        sub = self.subs[current_sub]
        if sub.status == 'select':
            sub.status = 'reject'
        elif sub.status == 'reject':
            sub.status = 'select'

        self.update_stack_scroller()    # updates colours
        self.stack_changed()            # will check if sub or stack


    def check_for_change(self):
        ''' whenever there are changes to the stack, check whether 
            anything needs saving
        '''

        # empty stack implies no changed
        if self.is_empty():
            self.changed = ''

        # stack contains calibration subs, so always allow master to be saved
        elif self.subs[0].sub_type in {'flat', 'dark', 'bias'}:
            self.changed = 'potential new master'

        # previous object
        elif Component.get('ObjectIO').existing_object:
            if {s.name.lower() for s in self.subs if s.status == 'reject'} != self.orig_rejects:
                # used to have this too: self.subs[0].exposure != self.original_exposure
                self.changed = 'sub select/reject status modified'
            else:
                self.changed = ''

        # live object, stack non-empty, therefore changed
        else:
            self.changed = 'new object'


    def stack_changed(self):
        ''' Called whenever we might need to update the stack 
            e.g. sub added/selected
        '''

        self.check_for_change()

        if self.is_empty() or not self.viewing_stack:
            return

        Component.get('MultiSpectral').stack_changed()

        self.update_status()


    def update_stack_scroller(self, *args):
        # sub_labels is array representing the stack
        # these are sub positions on the screen

        self.update_status()
        gui = self.app.gui.gui
        # can happen
        if not 'sub_0' in gui:
            return
        lab = gui['sub_0']['widget']
        i = self.selected_sub
        if i >= 0:
            lab.text = str(i + 1)
            lab.color = self.sub_colors[self.subs[i].status]
            lab.background_color = Component.get('FilterChooser').filter_properties.get(
                self.subs[i].filter,'?')['bg_color']
        else:
            lab.text = ' '
 

    def increment_sub(self, dt):
        if self.selected_sub == len(self.subs) - 1:
            self.selected_sub = 0
        else:
            self.selected_sub += 1


    def get_screen(self, dt=None):
        im = Component.get('Snapshotter').snap(return_image=True).copy()
        self.ims.append(im.convert('L'))
        if len(self.ims) < len(self.subs):
            self.selected_sub += 1
            Clock.schedule_once(self.get_screen, 0)
        else:
            save_path = '{:} {:}.gif'.format(
                    os.path.join(self.app.get_path('snapshots'), 
                        purify_name(Component.get('DSO').Name)), 
                    datetime.now().strftime('%d%b%y_%H_%M_%S'))
            self.ims[0].save(save_path, 
                save_all=True, 
                append_images=self.ims[1:], 
                optimize=True, 
                # include_color_table=False,
                duration=1000 / self.speed, 
                loop=0)
            # Component.get('Snapshotter').info('saved animation')
            toast(f'saved animation to {save_path}', duration=2)


    def make_animated_gif(self):
        # stop existing animation
        if hasattr(self, 'play_event'):
            Clock.unschedule(self.play_event)
        self.ims = []
        self.selected_sub = 0
        Clock.schedule_once(self.get_screen, 0)


    def sub_added(self):
        # called when new sub added to the stack by Watcher
        if self.selected_sub == len(self.subs) - 2:  # was on last sub before new one came in
            self.selected_sub = len(self.subs) - 1
        else:
            self.update_stack_scroller()


    def on_selected_sub(self, *args):
        # This is the key method that dictates whether the display has changed
        # on new sub, change of sub selection etc and hence stack updates

        s = self.selected_sub
        if (s < 0) or (s >= len(self.subs)):
            return

        if self.viewing_stack:
            self.stack_changed()
        else:
            ss = self.subs[s]
            Component.get('Monochrome').display_sub(
                ss.get_image(), 
                do_gradient=ss.sub_type=='light',
                fwhm=ss.fwhm if hasattr(ss, 'fwhm') else 0)

        self.update_stack_scroller()
        self.check_for_change()


    def get_current_displayed_image(self, first_sub=False):
        ''' Called by platesolver to get currently displayed image
             which might be short sub, sub, or stack
        '''
 
        if self.is_empty():
            ''' we have no image on the stack so must be faffing
                and pixel height therefore has to come from camera
            '''
            try:
                im = Component.get('Capture').last_faf
                if im is None:
                    logger.trace('last_faf is None')
                    return None
                self.pixel_height = Component.get('Camera').get_pixel_height()
                logger.trace(f'using faf with pixel height {self.pixel_height}')
            except:
                logger.trace('no stack and no short subs either')
                return None

        else:
            ''' we have stacked images
            '''
            if first_sub:
                im = self.subs[0].get_image()
            elif self.viewing_stack:
                im = self.get_stack()
            else:
                im = self.subs[self.selected_sub].get_image()

            # get pixel height from sub; not clear if we should
            # be mult by binning though
            ph = self.subs[0].pixel_height
            self.pixel_height = None if ph is None else ph * self.subs[0].binning

        return Component.get('View').do_flips(im)

               
    def get_pixel_height(self):
        if hasattr(self, 'pixel_height'):
            return self.pixel_height
        return None


    def get_RGB_at_pixel(self, x, y):
        ''' Used for G2V calibration: really need to do a proper stellar centroid
            around this pixel
        '''
        x = int(x)
        y = int(y)
        r, g, b = self.get_stack(filt='R'), self.get_stack(filt='G'), self.get_stack(filt='B')
        if r is None or g is None or b is None:
            return


    def get_centroids_for_platesolving(self):
        try:
            return self.subs[self.selected_sub].centroids
        except:
            logger.warning('no image for platesolving')
            return None


    def get_selected_sub_count(self):
        # used by calibrator
        return len([1 for s in self.subs if s.status == 'select'])


    def get_prop(self, prop=None, unique=True, average=False):
        ''' find value of specified property  (e.g. exposure time) from 
            subs on stack; if unique is set return None if more than one 
            value exists for the property
        '''
        if prop is None:
            return None
        pvals = {getattr(s, prop, None) for s in self.subs}
        if average:
            try:
                return np.mean(list(pvals))
            except:
                return None
        if len(pvals) == 0 or None in pvals:
            return None
        if unique:
            return getattr(self.subs[0], prop) if len(pvals) == 1 else None
        return pvals


    def get_stack(self, filt='all', calibration=False):
        ''' Return stack of selected subs of the specified filter. Caches
            results to prevent expensive recomputes. Does fast stack
            combination for addition or removal of a single sub assuming
            combination method is 'mean'. For calibration, uses all subs.
        '''

        # we have no subs
        if not self.subs:
            return None

        orig_sub_map = {s.name: s for s in self.subs}

        # subs up to current selected sub, except in case of calibration
        if calibration:
            subs = self.subs
        else:
            subs = self.subs[:self.selected_sub+1]

        # if no filter, choose all subs, otherwise restrict
        if filt == 'all':
            stk = [s for s in subs if s.status=='select']
        else:
            stk = [s for s in subs if (s.status=='select') & (s.filter==filt)]

        # none remain
        if not stk:
            return None

        return Component.get('StackCombiner').combine(
            stk, 
            orig_sub_map,
            filt=filt, 
            calibration=calibration)


    @logger.catch()
    def add_sub(self, sub):
        ''' called by ObjectIO.new_sub
        '''

        ''' check if sub is the same dims as existing stack, otherwise don't add; note
            dims are in reverse order to those of the numpy array
        ''' 
        if not self.is_empty():
            width, height = self.subs[0].image.shape
            if sub.shape[1] != width or sub.shape[0] != height:
                msg = f'sub dims {width} x {height} incompatible with current stack {self.subs[0].get_image().shape}'
                toast(msg)
                logger.error(msg)
                return

        self.process(sub)
        self.subs.append(sub)
        self.sub_added()

        # Add stack number
        sub.stack_num = len(self.subs)

        if sub.temperature is not None:
            Component.get('Session').temperature = sub.temperature


    def realign(self, *args):
        self.recompute(realign=True)


    def recompute(self, widgy=None, realign=False):
        # initial load, recompute or realign
        Component.get('Aligner').reset()
        Component.get('StackCombiner').reset()
        # self.stack_cache = {}
        initial_filters = {'B', 'H', 'S', 'O'}
        # shuffle-based realign
        if realign:
            np.random.shuffle(self.subs)
            # if we have a B or Ha, keep shuffling until we start with that
            if len([s for s in self.subs if s.filter in initial_filters]) > 0:
                while self.subs[0].filter not in initial_filters:
                    np.random.shuffle(self.subs)

            # Renumber the subs
            for num, sub in enumerate(self.subs):
                sub.stack_num = num + 1

        self.selected_sub = -1
        self.reprocess_event = Clock.schedule_once(self._reprocess_sub, 0)


    def _reprocess_sub(self, dt):
        if self.selected_sub + 1 < len(self.subs):
            self.process(self.subs[self.selected_sub + 1])
            self.selected_sub += 1
            self.reprocess_event = Clock.schedule_once(self._reprocess_sub, 0)
        else:
            self.reprocess_event = None
            self.stack_changed()


    def stop_load(self):
        if hasattr(self, 'reprocess_event') and self.reprocess_event is not None:
            Clock.unschedule(self.reprocess_event)


    def process(self, sub):
        # Process sub on recompute/realign

        sub.image = None  # force reload
        # sub.image = Component.get('BadPixelMap').remove_hotpix(sub.get_image())
        Component.get('BadPixelMap').process_bpm(sub)
        if sub.sub_type == 'light':
            Component.get('Calibrator').calibrate(sub)
            Component.get('Aligner').align(sub)
        elif sub.sub_type == 'flat':
            Component.get('Calibrator').calibrate_flat(sub)


    ''' Below, code for viewing/editing of FITs information in subs
    '''

    def show_subs_table(self, *args):
        ''' Called when user clicks on GUI
        '''

        if self.is_empty():
            return

        # build subs_table_records dict
        self.update_sub_records()

        # build table if necessary
        if not hasattr(self, 'subs_table'):
            self.subs_table = self.build_substable()
        
        self.subs_table.data = self.subs_table_records

        self.app.showing = 'subs'

        # check for redraw
        if self.subs_table not in self.app.gui.children:
            self.app.gui.add_widget(self.subs_table, index=0)

        # ensure that selected set aligns with subs display in GUI
        selected_set = set({s.name for s in self.subs if s.status == 'select'})
        self.subs_table.select(selected_set)
        self.subs_table.update()
        self.subs_table.show()    


    def build_substable(self):
        ''' Contruct table from stack
        '''

        cols={
            'stack_num': {'w': 45, 'label': 'N', 'type':int}, 
            'status': {'w': 65, 'label': 'Status'},
            'aligned': {'w': 70, 'label': 'Aligned'},
            'name': {'w': 200, 'align': 'left', 'label': 'Name', 'action': self.view_single_sub}, 
            'pp_create_time': {'w': 140, 'align': 'left', 'label': 'Date', 'sort': {'DateFormat': date_time_format}}, 
            'exposure': {'w': 60, 'label': 'Expo', 'type': float},
            'temperature': {'w': 60, 'label': 'Temp', 'type': float, 'display_fn': lambda x: f'{x:.1f}'},
            'gain': {'w': 60, 'label': 'Gain', 'type': float, 'display_fn': lambda x: f'{x:.0f}'},
            'offset': {'w': 55, 'label': 'Offset', 'type': float, 'display_fn': lambda x: f'{x:.0f}'},
            'binning': {'w': 40, 'label': 'Bin'},
            'filter': {'w': 40, 'label': 'Filt'},
            'fwhm': {'w': 80, 'label': 'FWHM', 'type': float, 'display_fn': lambda x: f'{x:.1f} px'},
            'minval': {'w': 70, 'label': 'Min', 'type': float, 'display_fn': lambda x: f'{x:.1f}%'},
            'maxval': {'w': 70, 'label': 'Max', 'type': float, 'display_fn': lambda x: f'{x:.1f}%'},
            'meanval': {'w': 70, 'label': 'Mean', 'type': float, 'display_fn': lambda x: f'{x:.1f}%'},
            'overexp': {'w': 70, 'label': 'Over', 'type': float, 'display_fn': lambda x: f'{x:.1f}%'}
            }

        # set up editable fields
        for c in ['exposure', 'temperature', 'gain', 'offset', 'binning', 'filter']:
            cols[c]['action'] = self.edit_prop

        return Table(
            size=Window.size,
            data=self.subs_table_records,
            name='Subs',
            description='Subs',
            cols=cols,
            on_hide_method=self.update_stack_from_table,
            initial_sort_column='stack_num'
            )


    def edit_prop(self, row, prop, value):
        # placeholder for now
        toast(f'editing prop {prop} {value} (not yet implemented)')


    def update_stack_from_table(self, *args):
        ''' change selected status based on checkboxes
        ''' 
        if len(self.subs_table.selected) == 0:
            return

        for sub in self.subs:
            sub.status = 'select' if sub.name in self.subs_table.selected else 'reject'

        self.update_stack_scroller()    # updates colours
        self.stack_changed()


    def update_sub_records(self):
        props = [
            'stack_num', 'meanval', 'minval', 'maxval', 'pp_create_time', 
            'exposure', 'filter', 'name', 'fwhm', 'gain', 'offset', 
            'temperature', 'binning', 'overexp', 'status', 'aligned'
            ]
        self.subs_table_records = {}
        for s in self.subs:
            props = {p: handleNAN(getattr(s, p) if hasattr(s, p) else 0) for p in props}
            self.subs_table_records[s.name] = props



    ''' This is potential machinery for editing; not currently used
    '''


    # def is_valid(self, field, value):
    #     if field == 'sub type':
    #         return value in {'dark', 'light', 'bias', 'flat'}
    #     if field in {'exposure', 'gain', 'offset'}:
    #         try:
    #             float(value)
    #             return True
    #         except Exception as e:
    #             return False
    #     if field == 'temperature':
    #         try:
    #             v = float(value)
    #             return v > -100 and v < 100
    #         except Exception as e:
    #             return False
    #     return False

    # def update_fits_header(self, path, key, value, comment=None):
    #     ''' modify headers
    #     '''
    #     # print('updating fits header', path, key, value)
    #     try:
    #         with fits.open(path, mode='update') as hdu:
    #             hdr = hdu[0].header
    #             if comment is None:
    #                 hdr[key] = value
    #             else:
    #                 hdr[key] = (value, comment)
    #             hdu.close()
    #     except Exception as e:
    #         logger.exception('Unable to update fits header for {:} ({:})'.format(path, e))
    #         return

    # def edit_value(self, *args):

    #     # validate value
    #     value = self.edited_value.text.strip()
    #     field = self.spinner.text

    #     if not self.is_valid(field, value):
    #         self.edited_value.foreground_color = (1, 0, 0, 1)
    #         return
    #     self.edited_value.foreground_color = (1, 1, 1, 1)

    #     # for each sub that is selected
    #     for selected_sub in self.subs_table.selected:
    #         sub = None
    #         for s in self.subs:
    #             if s.name == selected_sub:
    #                 sub = s
    #                 break
    #         if field == 'exposure':
    #             self.update_fits_header(sub.path, 'EXPOSURE', value, 'seconds')
    #             self.update_fits_header(sub.path, 'EXPTIME', value, 'seconds')
    #             self.update_fits_header(sub.path, 'EXPO', value, 'seconds')
    #             self.update_fits_header(sub.path, 'EXP', value, 'seconds')
    #         elif field == 'sub type':
    #             self.update_fits_header(sub.path, 'SUBTYPE', value)
    #             self.update_fits_header(sub.path, 'IMAGETYP', value)
    #         elif field == 'temperature':
    #             self.update_fits_header(sub.path, 'TEMPERAT', value)
    #             self.update_fits_header(sub.path, 'TEMP', value)
    #             self.update_fits_header(sub.path, 'CCD-TEMP', value)


    ''' Table for a single sub.

        might be easier to popup a box on hover or on clicking sub name 
        rather than having its own table; 
        on the other hand, perhaps we can edit properties this way too
    '''

    def build_singlesubtable(self):
        ''' Contruct table from stack
        '''

        self.single_sub = {}

        cols={
            'row': {'w': 80, 'align': 'left', 'label': 'line', 'type': int}, 
            'key': {'w': 120, 'align': 'left', 'label': 'Key'}, 
            'value': {'w': .5, 'align': 'left', 'label': 'Value'}, 
            'comment': {'w': .5, 'align': 'left', 'label': 'Comment'}, 
            }

        return Table(
            size=Window.size,
            data=self.single_sub,
            name='SingleSub',
            description='Single Sub',
            cols=cols,
            actions={},
            on_hide_method=self.app.table_hiding,
            initial_sort_column='row' 
            )


    def view_single_sub(self, row, *args):
        sub_name = str(row.key)
        self.subs_table.hide()
        if not hasattr(self, 'singlesub_table'):
            self.singlesub_table = self.build_singlesubtable()

        # find which sub it is
        sub = None
        for s in self.subs:
            if s.name == sub_name:
                sub = s
                break

        if sub is None:
            logger.warning(f'cannot find sub {sub_name}')
            return

        # load FITs
        try:
            with fits.open(sub.path, ignore_missing_end=True) as hdu1:
                hdu1.verify('silentfix')
                hdr = hdu1[0].header
        except Exception as e:
            logger.warning(f'cannot load FITs {sub.path} ({e})')
            return

        self.single_sub = {}
        lines = hdr.tostring(sep='\\n').split('\\n')
        for i, l in enumerate(lines[:-1]):
            kk = i + 1
            ll = l.split('=')
            if len(ll) == 1:
                ll = ll[0]
                if ll.strip().startswith('COMMENT'):
                    key = 'COMMENT'
                    rest = ll.strip()[8:]
                else:
                    key = ''
                    rest = ll
            elif len(ll) == 2:
                key, rest = ll[0], ll[1]
            else:
                key = ll[0]
                rest = '='.join(ll[1:])

            rr = rest.split('/')
            value = rr[0]
            if len(rr) == 1:
                comment = ''
            elif len(rr) == 2:
                comment = rr[1]
            else:
                comment = '/'.join(rr[1:]) 
            self.single_sub[kk] = {
                'row': kk, 
                'key': key.strip(), 
                'value': handleNAN(value.replace('"','').replace("'",'').strip()), 
                'comment': comment.strip()}

        self.singlesub_table.data = self.single_sub
        self.app.showing = 'singlesub'

        # check for redraw
        if self.singlesub_table not in self.app.gui.children:
            self.app.gui.add_widget(self.singlesub_table, index=0)

        self.singlesub_table.show()    





''' Camera component. All work is done via camera subclasses.
'''

from kivy.event import EventDispatcher

from jocular.component import Component
from jocular.device import DeviceFamily
# from jocular.settingsmanager import JSettings

from jocular.cameras.watchedcamera import WatchedCamera
from jocular.cameras.sxcamera import SXCamera
from jocular.cameras.ultrastar import SXUltrastarCamera
# from jocular.cameras.ascomcamera import ASCOMCamera
from jocular.cameras.simulatorcamera import SimulatorCamera
from jocular.cameras.asicamera import ASICamera

class Camera(Component, EventDispatcher, DeviceFamily):

	modes ={ 
		'Watched dir': 'WatchedCamera', 
		'Simulator': 'SimulatorCamera',
		#'ASCOM': 'ASCOMCamera',
		'ASI': 'ASICamera',
		'SX Lodestar': 'SXCamera',
		'SX Ultrastar': 'SXUltrastarCamera'
	}

	configurables = []

	default_mode = 'Watched dir'
	family = 'Camera'


	def on_new_object(self, *args):
		if self.connected():
			self.device.on_new_object()


	def on_previous_object(self, *args):
		if self.connected():
			self.device.on_previous_object()


	def capture_sub(self, **kwargs):
		if self.connected():
			return self.device.capture(**kwargs)


	def stop_capture(self):
		if self.connected():
			self.device.stop_capture()


	def get_image(self):
		if self.connected():
			return self.device.get_image()


	def get_pixel_height(self):
		if self.connected():
			return self.device.get_pixel_height()


	def get_sensor_temperature(self):
		if self.connected():
			return self.device.get_sensor_temperature()


	def set_ROI(self, ROI):
		if self.connected():
			self.device.set_ROI(ROI)


	def get_capture_props(self):
		if self.connected():
			return self.device.get_capture_props()


	def on_close(self, *args):
		if self.connected():
			self.disconnect()	



''' All the experimental stuff (mainly sliders) goes in this panel
'''

from kivy.app import App
from kivy.metrics import dp
from kivy.uix.gridlayout import GridLayout
from kivy.uix.anchorlayout import AnchorLayout
from kivy.properties import NumericProperty

from jocular.component import Component
from jocular.widgets.widgets import JSlider, LabelR
from jocular.panel import Panel


class Experimental(Panel, Component):

    fracbin = NumericProperty(1)


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.build()
        self.panel_opacity = 0


    def on_show(self):
        pass


    def build(self, *args):

        content = self.contents
        layout = AnchorLayout(anchor_x='center', anchor_y='bottom', size_hint=(1, 1))
        content.add_widget(layout)
        gl = GridLayout(size_hint=(None, None), cols=2, 
            width=dp(400), height=dp(30), spacing=(dp(5), dp(5)))
        gl.add_widget(LabelR(text='fractional binning'))
        slider = JSlider(
            size_hint=(None, None), width=dp(200), height=dp(30), 
            step=0.1, min=1, max=3, value=self.fracbin)
        slider.bind(value=self.fracbin_changed)
        gl.add_widget(slider)
        layout.add_widget(gl)
        self.app.gui.add_widget(self)


    def fracbin_changed(self, widget, *args):
        Component.get('Monochrome').fracbin = widget.value




from jocular.device import Device
from kivy.properties import StringProperty
#from jocular.settingsmanager import JSettings

#class GenericCamera(JSettings, Device):
class GenericCamera(Device):

	family = StringProperty('Camera')

	def on_new_object(self):
		self.last_capture = None

	def on_previous_object(self):
		self.last_capture = None

	def reset(self):
		self.last_capture = None

	def capture(self, **kwargs):
		return None

	def stop_capture(self):
		pass

	def set_ROI(self, ROI):
		pass

	def get_pixel_height(self):
		return None

	def get_sensor_temperature(self):
		return None

	def get_image(self):
		if hasattr(self, 'last_capture'):
			return self.last_capture
		return None

	def get_capture_props(self):
		''' Overridden if cam has props that might be useful to caller
			e.g. to write into the FITs header
		'''
		return None



''' Help panel, currently used for tooltips only, but could be
    extended to provide more help
'''


from kivy.app import App
from kivymd.uix.behaviors import HoverBehavior
from kivymd.uix.boxlayout import MDBoxLayout
from kivy.properties import BooleanProperty, StringProperty 
from kivy.lang import Builder

from jocular.component import Component


Builder.load_string(
'''
<Help>:
    helplabel: _label
    MDLabel:
        # canvas.before:
        #     Color:
        #         rgb: 1, 0, 0
        #         a: .1
        #     Rectangle:
        #         pos: self.pos
        #         size: self.size
        id: _label
        halign : 'center'
        font_style: 'H6'
''')


class TooltipBehavior(HoverBehavior):

    tooltip_text = StringProperty('')

    def on_enter(self, *args):
        Component.get('Help').show_tooltip(self.tooltip_text)

    def on_leave(self, *args):
        Component.get('Help').show_tooltip('')


class Help(MDBoxLayout, Component):

    show_tooltips = BooleanProperty(False)


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        App.get_running_app().gui.add_widget(self)
        Component.get('Appearance').show_tooltips = False


    def on_show_tooltips(self, *args):
        if not self.show_tooltips:
            self.helplabel.text  = ''


    def show_tooltip(self, text):
        if self.show_tooltips:
            self.helplabel.text = text




""" Appearance (font sizes, colours etc)
"""

from kivy.app import App
from kivy.properties import NumericProperty, StringProperty, BooleanProperty
from jocular.component import Component
from jocular.settingsmanager import JSettings


def sat_to_hue(val):
    hues = ["50", "100", "200", "300", "400", "500", "600", "700", "800", "900"]
    return hues[int((len(hues) - 1) * val / 100)]


class Appearance(Component, JSettings):

    highlight_color = StringProperty("BlueGray")
    lowlight_color = NumericProperty(50)
    lever_color = NumericProperty(30)
    hint_color = NumericProperty(25)
    ring_font_size = NumericProperty(14)
    form_font_size = NumericProperty(15)
    transparency = NumericProperty(0)
    colour_saturation = NumericProperty(60)
    show_tooltips = BooleanProperty(False)

    configurables = [
        (
            "highlight_color",
            {
                "name": "highlight colour",
                "options": [
                    "Red",
                    "Pink",
                    "Purple",
                    "DeepPurple",
                    "Indigo",
                    "Blue",
                    "LightBlue",
                    "Cyan",
                    "Teal",
                    "Green",
                    "LightGreen",
                    "Lime",
                    "Yellow",
                    "Amber",
                    "Orange",
                    "DeepOrange",
                    "Brown",
                    "Gray",
                    "BlueGray",
                ],
                "help": "choose a colour for highlights e.g. selected options",
            },
        ),
        (
            "colour_saturation",
            {
                "name": "colour saturation",
                "float": (0, 100, 1),
                "fmt": "{:.0f} percent",
            },
        ),
        (
            "lowlight_color",
            {
                "name": "icon/text grey level",
                "float": (0, 100, 1),
                "fmt": "{:.0f} percent",
                "help": "Gray level for normal text/icons",
            },
        ),
        (
            "lever_color",
            {
                "name": "lever grey level",
                "float": (0, 100, 1),
                "fmt": "{:.0f} percent",
                "help": "Gray level for levers (circular scrollers)",
            },
        ),
        (
            "hint_color",
            {
                "name": "hint text grey level",
                "float": (0, 100, 1),
                "fmt": "{:.0f} percent",
                "help": "Hint text appears above text entry boxes",
            },
        ),
        (
            "transparency",
            {
                "name": "ring transparency",
                "float": (0, 100, 1),
                "fmt": "{:.0f} percent",
                "help": "",
            },
        ),
        (
            "ring_font_size",
            {
                "name": "ring font size",
                "float": (10, 24, 1),
                "help": "Font size for labels used on the ring",
                "fmt": "{:.0f} points",
            },
        ),
        (
            "form_font_size",
            {
                "name": "form font size",
                "float": (10, 24, 1),
                "help": "Font size for form elements eg in DSO entry or session",
                "fmt": "{:.0f} points",
            },
        ),
        (
            "show_tooltips",
            {
                "name": "show tooltips",
                "switch": "",
            },
        ),
    ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()


    def on_ring_font_size(self, *args):
        self.app.ring_font_size = f"{self.ring_font_size:.0f}sp"


    def on_form_font_size(self, *args):
        self.app.form_font_size = f"{self.form_font_size:.0f}sp"


    def on_highlight_color(self, *args):
        self.app.theme_cls.accent_palette = self.highlight_color
        self.app.theme_cls.primary_palette = self.highlight_color


    def on_colour_saturation(self, *args):
        self.app.theme_cls.accent_hue = sat_to_hue(self.colour_saturation)
        self.app.theme_cls.primary_hue = sat_to_hue(self.colour_saturation)


    def on_lowlight_color(self, *args):
        lg = max(20, min(int(self.lowlight_color), 100)) / 100
        self.app.lowlight_color = [lg, lg, lg, 1]


    def on_lever_color(self, *args):
        lg = max(20, min(int(self.lever_color), 100)) / 100
        self.app.lever_color = [lg, lg, lg, 1]


    def on_hint_color(self, *args):
        lg = max(0, min(int(self.hint_color), 100)) / 100
        self.app.hint_color = [lg, lg, lg, 1]


    def on_transparency(self, *args):
        self.app.transparency = self.transparency / 100


    def on_show_tooltips(self, *args):
        Component.get('Help').show_tooltips = self.show_tooltips



''' Handles capture script menu and script generators.

    Important. Capture scripts are only altered by the user and
    are not changed in response to either new subs coming from
    the watcher, or from previous captures. Thus any subs with
    strange exposures, for example, don't change the capture
    script values. On the other hand, the GUI components for
    exposure, filter and script *are* updated to reflect the 
    most recent sub regardless of where it originates.
'''

import json
from loguru import logger

from kivy.app import App
from kivy.clock import Clock
from kivy.properties import OptionProperty
from kivy.metrics import dp
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.gridlayout import GridLayout

from jocular.component import Component
from jocular.exposurechooser import exp_to_str
from jocular.widgets.widgets import JMDToggleButton
from jocular.panel import Panel

faf_scripts = ['align', 'focus', 'frame']
light_scripts = ['light', 'seq']
calibration_scripts = ['dark', 'bias', 'flat', 'autoflat']
all_scripts = faf_scripts + light_scripts + calibration_scripts


class CaptureScript(Panel, Component):

    current_script = OptionProperty('align', options=all_scripts)
    capture_controls = {'devices', 'script_button', 'exposure_button', 'filter_button'}

    tooltips = {
        'align': 'brings up a reticle for initial alignment',
        'focus': 'typically longer exposure e.g. for use of focussing mask',
        'frame': 'framing subs; also useful for platesolving',
        'light': 'normal light subs in a single filter',
        'seq' : 'sequence of subs through more than one filter e.g. used for LRGB',
        'dark': 'create master dark and add to calibration library',
        'bias': 'create master bias and add to calibration library',
        'flat': 'create master flat and add to calibration library',
        'autoflat': 'as flat but exposure is estimated automatically (recommended)'
    }


    def __init__(self, **args):
        super().__init__(**args)
        self.app = App.get_running_app()
        try:
            with open(self.app.get_path('capture_scripts.json'), 'r') as f:
                self.scripts = json.load(f)
        except:
            # any problem loading => set up afresh
            self.scripts = {}
            for s in all_scripts:
                self.scripts[s] = {
                    'exposure': .1 if s in faf_scripts else 1, 
                    'filter': ['dark' if s in {'bias', 'dark'} else 'L']
                    }
            self.scripts['bias']['exposure'] = .001
            self.scripts['seq']['nsubs'] = 4
            self.scripts['seq']['filter'] = ['B', 'G', 'R', 'L']
            self.save()

        self.build()
        self.panel_opacity = 0

        # initialise via current script once dependencies have been built
        Clock.schedule_once(self.on_current_script, 0)


    def show_menu(self, *args):
        ''' display choice of next script, taking into account what is
            compatible with current subs or not
        '''

        if Component.get('Stacker').is_empty():
            for v in self.script_buttons.values():
                v.disabled = False

        else:
            if self.current_script in calibration_scripts:
                for k in light_scripts + calibration_scripts:
                    self.script_buttons[k].disabled = True
                self.script_buttons[self.current_script].disabled = False

            elif self.current_script in light_scripts:
                for k in calibration_scripts:
                    self.script_buttons[k].disabled = True
      

    def _button(self, name):
        return JMDToggleButton(
                text=name, 
                group='scripts',
                font_size='16sp',
                # height=dp(30),
                tooltip_text=self.tooltips.get(name, ''),
                on_press=self.script_chosen)


    def build(self, *args):
        content = self.contents
        self.script_buttons = {s: self._button(s) for s in all_scripts}
        layout = AnchorLayout(anchor_x='center', anchor_y='bottom', size_hint=(1, 1))
        content.add_widget(layout)
        gl = GridLayout(size_hint=(1, None), cols=4, height=dp(180), spacing=(dp(5), dp(5)))
        for scripts in [faf_scripts, light_scripts, calibration_scripts]:
            for s in scripts:
                gl.add_widget(self.script_buttons[s])
            for i in range(4 - len(scripts)):
                b = self._button(' ')
                b.disabled = True
                # b.md_bg_color = 1, .45, .45, 0
                gl.add_widget(b)
        layout.add_widget(gl)
        self.app.gui.add_widget(self)


    def on_show(self):
        ''' display choice of next script, taking into account what is
            compatible with current subs or not
        '''

        for s, but in self.script_buttons.items():
            but.state = 'down' if s == self.current_script else 'normal'

        if Component.get('Stacker').is_empty():
            for v in self.script_buttons.values():
                v.disabled = False

        else:
            if self.current_script in calibration_scripts:
                for k in light_scripts + calibration_scripts:
                    self.script_buttons[k].disabled = True
                self.script_buttons[self.current_script].disabled = False

            elif self.current_script in light_scripts:
                for k in calibration_scripts:
                    self.script_buttons[k].disabled = True


    def script_chosen(self, item):
        ''' user has selected script, so update
        '''
        self.current_script = item.text


    def save(self, *args):
        ''' save capture scripts
        '''
        try:
            with open(self.app.get_path('capture_scripts.json'), 'w') as f:
                json.dump(self.scripts, f, indent=1)
        except Exception as e:
            logger.exception(f'Unable to save capture_scripts.json ({e})')


    def on_new_object(self, *args):
        ''' default to framing at start
        '''
        self.app.gui.enable(self.capture_controls)
        try:
            camera_mode = Component.get('Camera').settings['current_mode']
        except:
            # catch this exception because on first time usage we can't get to settings
            camera_mode = 'Watched dir'
        # don't go into framing mode unless connected to a camera
        if camera_mode in {'Watched dir', 'Simulator'}:
            self.current_script = 'light'
        else:
            self.current_script = 'frame'
        self.on_current_script()


    def on_previous_object(self, *args):
        ''' don't allow user to perform captures when loading previous object
        '''
        self.app.gui.disable(self.capture_controls)


    def on_current_script(self, *args):
        ''' carry out any special actions when certain scripts are selected
        '''

        logger.debug(f'Changed script to {self.current_script}')
        self.app.gui.set('script_button', self.current_script)
        self.update()
        if self.current_script == 'align':
            Component.get('View').fit_to_window(zero_orientation=False)
            self.prev_transp = Component.get('Appearance').transparency
            Component.get('Appearance').transparency = 100
            self.app.gui.set('show_reticle', True, update_property=True)
        else:
            Component.get('Appearance').transparency = \
                self.prev_transp if hasattr(self, 'prev_transp') else 0
            self.app.gui.set('show_reticle', False, update_property=True)
        Component.get('StackCombiner').combine_method = \
            '80' if self.current_script == 'flat' else 'mean'
        # self.app.gui.set('80' if self.current_script == 'flat' else 'mean' , 
        #     True, update_property=True)


    def filterwheel_changed(self):
        ''' when filterwheel changes we need to change the available
            filters in the capture scripts and therefore update the scripts
        '''        
        state = Component.get('FilterWheel').get_state()
        filts = list(state['filtermap'].keys())
        logger.debug(f'filters available in new filterwheel {filts}')
        default = ['L'] if 'L' in filts else [f for f in filts if f != '-']
        logger.debug(f'Default filter is {default}')
        if len(default) == 0:
            default = 'L'
        for k, v in self.scripts.items():
            v['filter'] = [f for f in v['filter'] if f in filts]
            if len(v['filter']) == 0:
                v['filter'] = default if k not in {'dark', 'bias'} else ['dark']
        logger.debug('scripts changed to accommodate new filterwheel')
        self.update()


    def update(self):
        ''' update panel, gui elements and save capture script details
        ''' 
        script = self.scripts[self.current_script]
        self.app.gui.set('exposure_button', exp_to_str(script['exposure']))
        self.app.gui.set('filter_button', ''.join(script['filter']))
        self.save()
        self.reset_generator()


    def get_exposure(self):
        ''' called by ExposureChooser, and also by Watcher to get 
            exposure in case user wishes to override
        '''
        return self.scripts[self.current_script]['exposure']


    def get_sub_type(self):
        ''' called by Watcher and Capture
        '''
        if self.current_script  == 'seq':
            return 'light'
        if self.current_script == 'autoflat':
            return 'flat'
        return self.current_script


    def get_filters(self):
        ''' called by FilterChooser
        '''
        return self.scripts[self.current_script]['filter']


    def get_nsubs(self):
        ''' called by FilterChooser
        '''
        return self.scripts[self.current_script].get('nsubs', 1)


    def exposure_changed(self, exposure):
        ''' called by exposure chooser
        '''
        self.scripts[self.current_script]['exposure'] = exposure
        self.update()


    def filter_changed(self, filt, nsubs=None):
        ''' called by filter chooser
        '''
        self.scripts[self.current_script]['filter'] = filt
        if nsubs is not None:
            self.scripts['seq']['nsubs'] = nsubs
        self.update()


    def set_external_details(self, filt=None, exposure=None, sub_type=None):
        ''' Display capture details on the interface. Used for previous captures 
            and watched captures.
        '''
        if type(filt) == list:
            filt = ''.join(filt)
        self.app.gui.set('exposure_button', '?' if exposure is None else exp_to_str(exposure))
        self.app.gui.set('filter_button', '?' if filt is None else filt)
        self.app.gui.set('script_button', '?' if sub_type is None else sub_type)


    def faffing(self):
        return self.current_script in faf_scripts

    ''' define scripts using generators
    '''

    def reset_generator(self):
        logger.debug(f'reset {self.current_script} generator')
        self.generator = getattr(self, f'{self.current_script}_generator')()


    def light_generator(self):
        script = self.scripts['light']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose long'


    def seq_generator(self):
        script = self.scripts['seq']
        yield 'set exposure', script['exposure']
        while True:
            for f in Component.get('FilterChooser').order_by_transmission(script['filter']):
                yield 'set filter', f
                for i in range(script['nsubs']):
                    yield 'expose long'


    def frame_generator(self):
        script = self.scripts['frame']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose short'


    def focus_generator(self):
        script = self.scripts['focus']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose short'


    def align_generator(self):
        script = self.scripts['align']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose short'


    def dark_generator(self):
        yield 'set filter', 'dark'
        yield 'set exposure', self.scripts['dark']['exposure']
        while True:
            yield 'expose long'


    def bias_generator(self):
        yield 'set filter', 'dark'
        yield 'set exposure', self.scripts['bias']['exposure']
        while True:
            yield 'expose bias'


    def autoflat_generator(self):
        yield 'set filter', self.scripts['autoflat']['filter'][0]
        yield 'autoflat'
        while True:
            yield 'expose long'


    def flat_generator(self):
        script = self.scripts['flat']
        yield 'set filter', script['filter'][0]
        yield 'set exposure', script['exposure']
        while True:
            yield 'expose long'




''' Basic class to draw reticle
'''

from kivy.properties import NumericProperty, BooleanProperty, ObjectProperty
from kivy.lang import Builder
from kivy.uix.widget import Widget
from kivy.metrics import dp
from kivymd.uix.button import MDIconButton

from jocular.metrics import Metrics
from jocular.component import Component


# NB for kivy prior to RC 2 could use :set inner dp(10)
Builder.load_string('''

#:set inner 10

<Arrow>:
    text_color: app.theme_cls.primary_color
    theme_text_color: "Custom"
    icon: f'arrow-{root.direction}'
    x: root.reticle.x + root.xoffset - self.width / 2
    y: root.reticle.y + root.yoffset - self.height / 2 if root.reticle.show and root.reticle.mount else -1000

<Reticle>:
    id: _reticle
    canvas:
        Color: 
            rgba: app.theme_cls.primary_color if root.show else self.disabled_color
        Line:
            circle: self.x, self.y, inner, 0, 360, 100
        Line:
            circle: self.x, self.y, 3 * inner, 0, 360, 100
        Line:
            circle: self.x, self.y, 6 * inner, 0, 360, 100
        Line:
            points: self.x, self.y - inner, self.x, self.y - 2 * self.radius 
        Line:
            points: self.x, self.y + inner, self.x, self.y + 2 * self.radius 
        Line:
            points: self.x - 2 * self.radius , self.y, self.x - inner, self.y
        Line:
            points: self.x + 2 * self.radius , self.y, self.x + inner, self.y
    disabled_color: [0, 0, 0, 0]

''')

class Arrow(MDIconButton):

    xoffset = NumericProperty(0)
    yoffset = NumericProperty(0)
    reticle = ObjectProperty(None)

    def __init__(self, direction='down', rate=1, reticle=None, **kwargs):

        self.direction = direction
        self.rate = rate
        self.reticle = reticle
        dist = dp(20 + self.rate * 80)
        if direction == 'left':
            self.xoffset = -dist
        elif direction == 'right':
            self.xoffset = dist
        elif direction == 'up':
            self.yoffset = dist
        elif direction == 'down':
            self.yoffset = -dist
        self.user_font_size = f'{20 + 7*rate}sp'
        super().__init__(**kwargs)


class Reticle(Component, Widget):

    radius = NumericProperty(1)
    show = BooleanProperty(False)
    mount = BooleanProperty(False)

    def __init__(self, **kwargs):
        self.pos = Metrics.get('origin')
        self.radius = Metrics.get('ring_radius')['image']
        super().__init__(**kwargs)
        # don't add arrows
        # until mount control working
        # for direction in ['left', 'right', 'up', 'down']:
        #     for rate in [1, 2, 3]:
        #         arrow = Arrow(
        #             direction=direction, rate=rate, reticle=self,
        #             on_press=partial(self.move, direction, rate),
        #             on_release=self.stop_moving)
        #         self.add_widget(arrow)


    # use these if ever get mount control going
    
    # def on_show(self, *args):
    #     if self.show:
    #         # check if we have a mount connected
    #         self.mount = Component.get('Telescope').connected()

    # def move(self, direction, rate, widget):
    #     # supposedly rate is in degrees per second so reduce it as follows
    #     Component.get('Telescope').move(direction=direction, rate=rate)

    # def stop_moving(self, *args):
    #     Component.get('Telescope').stop_moving()



''' Native driver for Starlight Xpress cameras. Currently supports Lodestar
	only. Use zadig to install libusb driver on Windows.
'''

import time
import array
from concurrent.futures import ThreadPoolExecutor
import numpy as np
from functools import partial

from loguru import logger
from kivy.properties import BooleanProperty
from kivy.clock import Clock

from jocular.cameras.genericcamera import GenericCamera
from jocular.utils import toast

SX_CLEAR_PIXELS = 1
SX_READ_PIXELS_DELAYED = 2
SX_READ_PIXELS = 3
SX_RESET = 6

# convert int formats
def convert_int(x, from_type='uint16', to_type='uint8'):
	return np.array([x], dtype=from_type).view(dtype=to_type)

class SXCamera(GenericCamera):

	lodestar = BooleanProperty(True)

	def connect(self):
		''' Connect to camera. Currently supports Lodestar only.
		'''

		if self.connected:
			return

		self.connected = False
		self.sxcamera = None

		try:
			import usb.core
			import usb.util
		except:
			self.status = 'no usb.core or usb.util'
			return

		# try to find Lodestar
		self.lodestar = False
		try:
			self.sxcamera = usb.core.find(idVendor=0x1278, idProduct=0x0507)
			if self.sxcamera is None:
				self.status = 'cannot find Lodestar'
				logger.info(self.status)
				return
			self.lodestar = True
			logger.info('found Lodestar')
		except Exception as e:
			logger.exception(e)

		# if self.sxcamera is None:
		# 	self.status = 'cannot find Lodestar'
		# 	logger.info(self.status)
		# 	return

		try:
			self.sxcamera.get_active_configuration()
		except Exception as e:
			logger.exception('problem getting active configuration ({:})'.format(e))
			self.status = 'cannot get active configuration'
			self.clean_up()
			return

		self.connected = True

		# used in capture; note h // 2 for interlaced; needs changing for ultrastar no doubt
		if self.lodestar:
			self.width, self.height = 752, 580
		else:
			self.width, self.height = 1392, 1040

		self.half_height = self.height // 2
		self.lswidth, self.mswidth = convert_int(self.width)
		self.lsheight, self.msheight = convert_int(self.half_height)

		self.status = 'SX Lodestar: {:} x {:} pixels'.format(
			self.width, self.height)

	def disconnect(self):
		if self.connected:
			self.connected = False
			self.status = 'disconnected'

	def stop_capture(self):
		# Cancel any pending reads
		if hasattr(self, 'capture_event'):
			if not self.internal_timing:  
				# don't cancel reading if using internal timing (causes crash on LS)
				self.capture_event.cancel()

	def clean_up(self):
		logger.debug('cleaning up')
		if hasattr(self, 'lodestar'):
			import usb.util
			usb.util.dispose_resources(self.sxcamera)
			logger.debug('Camera: disposed')

	def handle_failure(self, message, exception=None):
		toast('SX camera problem: {:}'.format(message))
		if exception is not None:
			logger.exception('Camera: handle_failure {:} ({:})'.format(message, exception))
		self.connected = False   # to ensure reconnect if that was the issue
		self.last_capture = None
		try:
			self.clean_up()
		except Exception as e:
			logger.exception('Camera: problem disposing {:}'.format(e))
		finally:
			if self.on_failure is not None:
				self.on_failure()

	def capture(self, exposure=None, on_capture=None, on_failure=None,
		binning=None, internal_timing=False, return_image=False, is_bias=False,
		is_faf=False):

		# is_faf not used in this version

		if is_bias:
			self.capture_bias(on_capture=on_capture, on_failure=on_failure)
			return

		logger.debug('Capturing from SX')
		# store these so they can be used without argument passing in futures etc
		self.on_capture = on_capture
		self.on_failure = on_failure
		self.exposure = exposure
		self.binning = binning
		self.internal_timing = internal_timing
		self.return_image = return_image
		self.last_capture = None

		# this in orig
		# self.connect()

		if not self.connected:
			toast('SX camera is not connected')
			# self.handle_failure('camera not connected')
			return

		try:
			if internal_timing:
				logger.debug('SX internally-timed exposure {:3f}'.format(exposure))
				# do internal_timed exposure
				self.start_internal_exposure()
				if return_image:
					time.sleep(exposure)
					return self.sxcamera_read()
				else:
					self.capture_event = Clock.schedule_once(
						self.read_internal_exposure, max(0.05, exposure)
					)
			else:
				# externally-timed
				logger.debug('SX external timed exposure {:3f}'.format(exposure))
				self.start_external_exposure()
				if return_image:
					time.sleep(exposure)
					return self.read_external_exposure()
				else:
					self.capture_event = Clock.schedule_once(
						self.read_external_exposure, max(0.05, self.exposure)
					)

		except Exception as e:
			self.handle_failure('capture', e)

	def capture_bias(self, on_capture=None, on_failure=None):
		# special case for bias as it needs even and odd for internally-timed exposure
		self.on_capture = on_capture
		self.on_failure = on_failure
		try:
			expo = 0.001
			self.exposure_command(rows='odd', exposure=expo)
			time.sleep(expo)
			odd_pixels = self.sxcamera.read(0x82, self.height * self.width, 10000)
			self.exposure_command(rows='even', exposure=expo)
			time.sleep(expo)
			even_pixels = self.sxcamera.read(0x82, self.height * self.width, 10000)
			pool = ThreadPoolExecutor(3)
			future = pool.submit(partial(self.deinterlace, odd_pixels, even_pixels))  # thread handles read
			future.add_done_callback(self.image_ready)  # when future is done, call image_ready
		except Exception as e:
			self.handle_failure('capture_bias', e)

	def exposure_command(self, rows='odd', exposure=None):
		# generate write command for sensor
		logger.debug('starting exposure')
		try:
			rowcode = {'odd': 1, 'even': 2, 'all': 3}
			cmd = SX_READ_PIXELS if exposure is None else SX_READ_PIXELS_DELAYED
			params = array.array('B', [0x40, cmd, rowcode[rows], 0, 0, 0, 10, 0, 
				0, 0, 0, 0, self.lswidth, self.mswidth, self.lsheight, self.msheight, 1, 1])
			if exposure is not None:
				# add exposure details to command
				expo = int(exposure * 1000)  #  ms
				exp1, exp2, exp3, exp4 = convert_int(expo, from_type='uint32')
				params = params + array.array('B', [exp1, exp2, exp3, exp4])
			self.sxcamera.write(1, params, 1000)
		except Exception as e:
			self.handle_failure('exposure_command', e)

	def start_internal_exposure(self):
		try:
			if self.lodestar:
				self.exposure_command(rows='odd', exposure=self.exposure)
		except Exception as e:
			self.handle_failure('start_internal_exposure', e)

	def start_external_exposure(self):
		try:
			self.sxcamera.ctrl_transfer(0x40, SX_RESET, 0, 0, 0)
			self.sxcamera.ctrl_transfer(0x40, SX_CLEAR_PIXELS, 0, 0, 0)
			logger.debug('started external exposure')
		except Exception as e:
			self.handle_failure('start_external_exposure', e)

	def read_internal_exposure(self, *args):
		logger.debug('reading internal exposure')
		try:
			pool = ThreadPoolExecutor(3)
			future = pool.submit(self.sxcamera_read)  # thread handles read
			future.add_done_callback(self.image_ready)  # when future is done, call image_ready
		except Exception as e:
			self.handle_failure('read_internal_exposure', e)

	def read_external_exposure(self, *args):
		''' read sensor at end of exposure time
		'''
		try:
			if self.lodestar:
				self.exposure_command(rows='odd')
				odds = self.sxcamera.read(0x82, self.height * self.width, timeout=10000)
				self.exposure_command(rows='even')
				evens = self.sxcamera.read(0x82, self.height * self.width, timeout=10000)
				self.last_capture = self.deinterlace(odds, evens)

			if self.on_capture is not None:
				self.on_capture()
		except Exception as e:
			self.handle_failure('read_external_exposure', e)

	def sxcamera_read(self):
		''' This is really the 'fast' read used for framing etc as it reads 
			odd rows and duplicates them; not convinced it is necessary any
			more because deinterlacing is fast; we can then get rid of the
			special case for bias above
		'''
		try:
			if self.lodestar:
				logger.debug('getting odd pixels')
				odd_pixels = self.sxcamera.read(0x82, self.height * self.width, 10000)
				return self.deinterlace(odd_pixels, odd_pixels)
		except Exception as e:
			self.handle_failure('sxcamera_read', e)

	def image_ready(self, future):
		self.last_capture = future.result()  # store image read from sensor
		logger.debug('im stats min {:.4f} max {:.4f} mean {:.4f}'.format(
			np.min(self.last_capture), np.max(self.last_capture), np.mean(self.last_capture)))
		if self.on_capture is not None:
			Clock.schedule_once(self.on_capture, 0)

	def get_capture_props(self):
		''' Any specific info for caller
		'''
		return {
			'camera': 'Lodestar X2',
			'pixel_width': 8.6,
			'pixel_height': 8.3
			}

	def get_pixel_height(self):
		return 8.3

	def deinterlace(self, odd8, even8):
		''' De-interlace Lodestar.
		'''

		logger.debug('deinterlacing')
		# convert uint8 arrays to uint16 and reshape
		odd = np.frombuffer(odd8, dtype='uint16').reshape(self.half_height, self.width)
		even = np.frombuffer(even8, dtype='uint16').reshape(self.half_height, self.width)

		# generate new array (full height)
		pix = np.zeros((self.height, self.width))

		# insert odd rows
		pix[::2, :] = odd

		# insert even rows and normalise to account for slight delay in reading
		pix[1::2, :] = even * (np.mean(odd) / np.mean(even))

		return pix / 2 ** 16






''' Home to platesolvers. Currently just the one: fastmatch. 
'''


import numpy as np
from scipy.spatial.distance import cdist
from skimage.transform import estimate_transform
from loguru import logger

from jocular.processing.starextraction import extract_stars
from jocular.RA_and_Dec import RA, Dec
from jocular.uranography import (
    to360,
    make_tile, get_tiles, in_tile,
    Eq2Cart,
    spherical_distance,
    radec2pix, pix2radec
)

class PlatesolverException(Exception):
    pass


''' generalise fast match so it is not dependent on units of measurement; in particular,
    the proximity & match thresholds need to be altered for the case where star1 and star2
    are quite close (ie unit distance represents two close points)

    also, rename it pointcloud_match or similar
    might be other optimisations (might wish to pregenerate the combinations, but
    taking note of recomputation)

    also, magrange doesn't seem to be that useful and is too specific; if idea
    is to reduce the mag diff between star1 and star2 perhaps have a max depth instead
'''


def fastmatch(
    ref=None,
    im=None,
    mag_range=5,
    min_matches=8,
    proximity=0.05,
    match_threshold=None,
    target_matches=20
):
    ''' find possible match of image points 'im' to reference stars 'ref' ;
    both ref and im are ordered by magnitude; works by choosing a pair
    of image points, rotating and scaling so they sit at (0, 0) and (1, 0);
    then searching across pairs of image points at a similar separation, applying
    similar rotation/scaling, then counting matches; return as soon as we get
    max_matches, or search exhausted
    '''

    max_matches = 0
    best_matches = None
    kk = match_threshold ** 2
    n_im, n_ref = len(im), len(ref)
    cnt = 0

    for i1 in range(0, n_im):
        # normalise image based on star index i1
        im1 = im - im[i1, :]
        for i2 in [i for i in range(i1 + 1, min(n_im, i1 + mag_range)) if i != i1]:
            # rotate and scale so i2 is at (1, 0)
            cos, sin = im1[i2, 0], -im1[i2, 1]
            d = cos ** 2 + sin ** 2
            im2 = np.dot(im1, np.array([[cos, -sin], [sin, cos]]).T) / d
            # match threshold takes into account scaling by d
            # mt = (match_arcsec / (3600 * (d ** 0.5))) ** 2.0  # we match squared dist
            mt = kk / d  # we match squared dist
            min_x, min_y = np.min(im2, axis=0) - mt
            max_x, max_y = np.max(im2, axis=0) + mt
            for r1 in range(0, n_ref):
                ref1 = ref - ref[r1, :]
                for r2 in [
                    r for r in range(r1 + 1, min(n_ref, r1 + mag_range)) if r != r1
                ]:
                    cos, sin = ref1[r2, 0], -ref1[r2, 1]
                    d2 = cos ** 2 + sin ** 2
                    if np.abs(d - d2) < proximity:
                        ref2 = np.dot(ref1, np.array([[cos, -sin], [sin, cos]]).T) / d2
                        mind_x, mind_y = np.min(ref2, axis=0)
                        maxd_x, maxd_y = np.max(ref2, axis=0)
                        # don't check if anything outside range
                        if (
                            max_x < maxd_x
                            and max_y < maxd_y
                            and min_x > mind_x
                            and min_y > mind_y
                        ):
                            cnt += 1
                            dists = cdist(im2, ref2, metric='sqeuclidean')
                            matches = [
                                (i, j)
                                for i, j in enumerate(np.argmin(dists, axis=1))
                                if dists[i, j] < mt
                            ]
                            n_matches = len(matches)
                            if n_matches >= target_matches:
                                return matches
                            if n_matches > max_matches:
                                max_matches = n_matches
                                best_matches = matches
    # return matches
    logger.debug('count {:}'.format(cnt))
    return best_matches


def get_platesolving_stars(star_db, tile):
    # get stars in tile; star_db is a npz file indexed by
    # ra and dec

    ra_tiles, dec_tiles = get_tiles(tile, 30, 10)
    ras, decs, mags = np.array([]), np.array([]), np.array([])
    for ra in ra_tiles:
        for dec in dec_tiles:
            quad = 'r{:}_d{:}_'.format(ra, dec)
            ras = np.append(ras, star_db[quad+'ra'])
            decs = np.append(decs, star_db[quad+'dec'])
            mags = np.append(mags, star_db[quad+'mag']/100.) # mags were stored as ints * 100

    locs = in_tile(tile, ras, decs)
    return ras[locs], decs[locs], mags[locs]


def platesolve(
    path=None, 
    im=None, 
    ra0=None, 
    dec0=None, 
    star_db=None,
    nstars=30, 
    min_matches=10, 
    max_matches=10, 
    degrees_per_pixel=None,
    nrefs=80,
    match_arcsec=15, 
    mag_range=5, 
    star_method='photutils', 
    target_matches=25,
    verbose=False):

    ''' Attempt to solve image at around ra0, dec0
    '''

    if path is None and im is None:
        raise PlatesolverException('must supply either an image or a FITs file path')

    if path is not None:
        raise PlatesolverException('Reading fits not yet supported')

    if ra0 is None or dec0 is None:
        raise PlatesolverException('must supply ra0/dec0 for approx centre of field')

    # extract star centroids and fluxes
    if verbose:
        logger.info('extracting stars')

    # extract stars (uses photutils; could place this in starextraction library)
    stars = extract_stars(
        im, 
        star_method=star_method, 
        target_stars=nstars,
        nstars=nstars,
        binfac=2,
        reset_threshold=False)

    x, y, flux = stars['xcentroid'], stars['ycentroid'], stars['flux']
    fwhm = np.median(stars['fwhm'])
    
    if len(x) < min_matches:
        raise PlatesolverException('insufficient stars found (needs {:}, found {:})'.format(min_matches, len(x)))

    if verbose:
        logger.info('extracted {:} stars'.format(len(x)))

    h, w = im.shape
    fov = h * degrees_per_pixel

    # convert flux to mags
    mags = -2.5 * np.log10(flux / np.max(flux))

    if verbose:
        logger.debug('mag range {:.1f}'.format(np.max(mags)))

    # sort in descending order of brightness anbd limit to nstars
    inds = np.argsort(mags)[:nstars]
    x_im, y_im = x[inds], y[inds]

    im_stars = degrees_per_pixel * np.vstack([x_im, y_im]).T

    # get platesolving reference stars in sufficiently large FOV
    ras, decs, mags = get_platesolving_stars(star_db, make_tile(ra0, dec0, fov=2*fov))

    # sort refs by mag and limit to N-brightest
    inds = np.argsort(mags)[:nrefs]

    # convert to cartesian coords
    ras, decs = Eq2Cart(ras[inds], decs[inds], ra0, dec0)

    ref_stars = np.degrees(np.vstack([ras, decs]).T)

    # do fastmatch with ref and im stars in degrees
    matches = fastmatch(
        ref=ref_stars,
        im=im_stars,
        match_threshold=match_arcsec / 3600.,
        min_matches=min_matches,
        target_matches=target_matches,
        mag_range=mag_range,
        proximity=0.05
    )

    if matches is None:
        raise PlatesolverException('no match')
    if len(matches) < min_matches:
        raise PlatesolverException('insufficient match')
    
    logger.info('matched {:}'.format(len(matches)))

    # create coord pairs of matching stars  
    src = np.vstack([ras, decs]).T[[j for (i, j) in matches], :]
    dst = np.vstack([x_im, y_im]).T[[i for (i, j) in matches], :]

    # estimate transform with scale, rotation and translation
    projection = estimate_transform('affine', src, dst)

    # find centre of frame in RA/Dec
    ra_centre, dec_centre = pix2radec(
        w // 2, h // 2, ra0, dec0, projection)

    # compute where North is (TO DO: second line might not work at poles....)
    xx, yy = radec2pix(
        np.array([ra0, ra0]),
        np.array([dec0 - 0.5, dec0 + 0.5]),
        ra0, dec0,
        projection
    )
    rotation = np.degrees(np.arctan2(yy[1] - yy[0], xx[1] - xx[0]))
    north = to360(90 - rotation)

    fov_w = spherical_distance(
        *pix2radec(0, 0, ra0, dec0, projection), 
        *pix2radec(w, 0, ra0, dec0, projection)
        )
    fov_h = spherical_distance(
        *pix2radec(0, 0, ra0, dec0, projection), 
        *pix2radec(0, h, ra0, dec0, projection)
        )

    return {
        'tile_ra0': ra0,
        'tile_dec0': dec0,
        'ra_centre': ra_centre,
        'dec_centre': dec_centre,
        'ra': str(RA(ra_centre)),
        'dec': str(Dec(dec_centre)),
        'width': w,
        'height': h,
        'fov_w': fov_w,
        'fov_h': fov_h,
        'north': north,
        'projection': projection,
        'im_stars': im_stars,
        'ref_stars': ref_stars,
        'fwhm': fwhm
    }





import os

from functools import partial
from loguru import logger
from pathlib import Path

from kivy.metrics import dp
from kivy.app import App
from kivy.uix.spinner import Spinner
from kivy.uix.button import Button
from kivymd.uix.slider import MDSlider
from kivymd.uix.selectioncontrol import MDSwitch
from kivymd.uix.filemanager import MDFileManager

from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from jocular.widgets.widgets import LabelL, LabelR


@logger.catch()
def configurable_to_widget(text=None, name=None, spec=None, initval=None, 
	helptext=None, changed=None, textwidth=None, widgetwidth=None):
	''' creates a widget and its handler, which calls function 'changed' on any change
	'''

	if textwidth is None:
		textwidth = dp(330) # was 280
	if widgetwidth is None:
		widgetwidth = dp(140) # was 150

	font_size = '15sp'
	help_font_size = '13sp'
	color = App.get_running_app().theme_cls.accent_color

	bv = BoxLayout(padding=(dp(5), dp(5)), size_hint=(1, None), height=dp(55), 
		orientation='vertical')
	bh = BoxLayout(size_hint=(1, .55))
	# bh.add_widget(Label(size_hint=(1, 1)))
	bh.add_widget(LabelR(text=text, size_hint=(None, 1), width=textwidth, font_size=font_size))

	if 'options' in spec:
		opts = spec['options']
		widget = Spinner(text=initval, values=spec['options'], 
			size_hint=(None, 1), width=widgetwidth, font_size=font_size)
		widget.bind(text=partial(__option_changed, name, spec, changed))
		bh.add_widget(widget)
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth))

	elif 'boolean' in spec:
		opts = spec['boolean'].keys()
		# lookup names e.g. yes/no for true or false
		val = {v: k for k, v in spec['boolean'].items()}[initval]
		widget = Spinner(text=val, values=opts, size_hint=(None, 1), width=widgetwidth, 
			font_size=font_size)
		widget.bind(text=partial(__boolean_changed, name, spec, changed))
		bh.add_widget(widget)
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth))

	elif 'switch' in spec:
		widget = MDSwitch(size_hint=(None, 1), width=widgetwidth, active=initval,
			pos_hint={'center_x': .5, 'center_y': .5})
		widget.bind(active=partial(__switch_changed, name, spec, changed))
		bh.add_widget(widget)
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth - dp(36)))
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth))

	elif 'float' in spec:
		fmt = spec.get('fmt', '{:.2f}')
		slabel = LabelL(text=fmt.format(initval), size_hint=(None, 1), width=widgetwidth, 
			font_size=font_size, color=color)
		bh.add_widget(slabel)
		smin, smax, step = spec['float']
		widget = MDSlider(size_hint=(None, 1), width=1.5*widgetwidth, 
			step=step, min=smin, max=smax, value=float(initval))
		widget.hint_bg_color=(.6, .6, .6, 1)
		# no need for this in kivymd 1.0.2
		# widget._set_colors()
		widget.bind(value=partial(__sfloat_changed, name, spec, slabel, fmt, changed))
		bh.add_widget(widget)

	elif 'double_slider_float' in spec:
		fmt = spec.get('fmt', '{:.2f}')
		slabel = LabelL(text=fmt.format(initval), size_hint=(None, 1), 
			width=widgetwidth, font_size=font_size, color=color)
		bh.add_widget(slabel)
		smin, smax = spec['double_slider_float']
		val1 = int(initval)
		val2 = abs(initval - val1)
		bh2 = BoxLayout(orientation='vertical', size_hint=(None, 1), width=1.5*widgetwidth)
		slider1 = MDSlider(size_hint=(1, .5), width=1.5*widgetwidth, 
			step=1, min=smin, max=smax, value=val1)
		slider2 = MDSlider(size_hint=(1, .5), width=1.5*widgetwidth, 
			step=.01, min=0, max=1, value=val2)
		bh2.add_widget(slider1)
		bh2.add_widget(slider2)
		slider1.bind(value=partial(
			__dfloat_changed, name, spec, slabel, fmt, changed, slider1, slider2))
		slider2.bind(value=partial(
			__dfloat_changed, name, spec, slabel, fmt, changed, slider1, slider2))
		bh.add_widget(bh2)

	elif 'action' in spec:
		widget = Button(text=spec['button'], size_hint=(None, 1), width=widgetwidth, 
			font_size=font_size)
		widget.bind(on_press=partial(__action_pressed, name, spec, changed))
		bh.add_widget(widget)
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth))

	elif 'filechooser' in spec:
		widget = Button(text='choose...', size_hint=(None, 1), width=widgetwidth,
		 font_size=font_size)
		widget.bind(on_press=partial(__filechooser_pressed, name, spec, changed, initval))
		bh.add_widget(widget)
		bh.add_widget(Label(size_hint=(None, 1), width=widgetwidth))

	# bh.add_widget(Label(size_hint=(1, 1)))

	# lower row contains only help text
	blow = BoxLayout(size_hint=(1, .45))
	blow.add_widget(Label(text=helptext, size_hint=(1, 1),
		font_size=help_font_size, color=(.5, .5, .5, 1)))
	bv.add_widget(bh)
	bv.add_widget(blow)

	return bv


def __sfloat_changed(name, spec, slabel, fmt, changed, slider, *args):
	slabel.text = fmt.format(slider.value)
	changed(name, slider.value, spec)


def __dfloat_changed(name, spec, slabel, fmt, changed, slider1, slider2, *args):
	val = slider1.value
	if val < 0:
		val -= slider2.value
	else:
		val += slider2.value
	slabel.text = fmt.format(val)
	changed(name, val, spec)


def __option_changed(name, spec, changed, widget, value, *args):
	changed(name, value, spec)


def __boolean_changed(name, spec, changed, widget, value, *args):
	changed(name, spec['boolean'][value], spec)


def __switch_changed(name, spec, changed, widget, *args):
	changed(name, widget.active, spec)


def __action_pressed(name, spec, changed, widget, *args):
	# call spec with widget also since we need to interact with it
	# to signal success/failure/state
	spec['widget'] = widget
	changed(name, spec['action'], spec)


def __filechooser_pressed(name, spec, changed, initpath, *args):
	fm = MDFileManager(search='dirs')
	fm.exit_manager = partial(exit_filemanager, fm)
	fm.select_path = partial(handle_selection, name, changed, spec, fm)
	# check if initpath is writeable and if not, bring up in home dir
	if initpath is not None:
		try:
			with open(os.path.join(initpath, '.written'), 'w') as f:
				f.write('can write')
		except:
			initpath = None
	if initpath is None:
		initpath = str(Path.home())
	# use show disks allows showing devices 
	# fm.show_disks()
	fm.show(initpath)


def exit_filemanager(widget, *args):
	widget.close()


def handle_selection(name, changed, spec, widget, path):
	changed(name, path, spec)
	widget.close()

